# -------------------------------------------------------------
#    Copyright 2004-2011 Synopsys, Inc.
#    Copyright 2010-2011 Mentor Graphics Corporation
#    Copyright 2010-2011 Cadence Design Systems, Inc.
#    Copyright 2019 Tuomas Poikela
#    All Rights Reserved Worldwide
#
#    Licensed under the Apache License, Version 2.0 (the
#    "License"); you may not use this file except in
#    compliance with the License.  You may obtain a copy of
#    the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in
#    writing, software distributed under the License is
#    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#    CONDITIONS OF ANY KIND, either express or implied.  See
#    the License for the specific language governing
#    permissions and limitations under the License.
#    TODO add modifications
# -------------------------------------------------------------

from ..base.sv import sv
from ..base.uvm_debug import uvm_debug
from ..base.uvm_object import UVMObject
from ..base.uvm_globals import UVM_LOW, UVM_MEDIUM
from ..macros.uvm_message_defines import (uvm_fatal, uvm_error, uvm_warning,
    uvm_info)
from ..macros.uvm_object_defines import uvm_object_utils
from .uvm_reg_model import (UVMRegMapAddrRange, UVM_BIG_ENDIAN, UVM_BIG_FIFO, UVM_FIELD, UVM_HIER,
                            UVM_LITTLE_ENDIAN, UVM_LITTLE_FIFO, UVM_MEM, UVM_NOT_OK, UVM_NO_HIER,
                            UVM_REG)
from .uvm_reg_item import UVMRegBusOp
from ..seq import UVMSequenceBase
from .uvm_reg_cbs import UVMRegReadOnlyCbs, UVMRegWriteOnlyCbs

UVM_VERB_MEM_MAP = UVM_LOW


class UVMRegMapInfo:
    def __init__(self):
        self.offset = 0
        self.rights = ""
        self.unmapped = False
        self.addr = []  # uvm_reg_addr_t[]
        self.frontdoor = None  # uvm_reg_frontdoor
        self.mem_range = UVMRegMapAddrRange()

        # if set marks the UVMRegMapInfo as initialized,
        # prevents using an uninitialized map (for instance if the model
        # has not been locked accidently and the maps have not been computed before)
        self.is_initialized = False

    def convert2string(self):
        res = ("offset: " + str(self.offset) + ', rights: ' + self.rights
            + ', unmapped: ' + str(self.unmapped) + ', addr: ' + str(self.addr)
            + ', range: ' + self.mem_range.convert2string())
        return res


class UVMRegTransactionOrderPolicy(UVMObject):
    def __init__(self, name='policy'):
        UVMObject.__init__(self, name)

    def order(self, q):
        """
        the order() function may reorder the sequence of bus transactions
        produced by a single uvm_reg transaction (read/write).
        This can be used in scenarios when the register width differs from
        the bus width and one register access results in a series of bus transactions.
        the first item (0) of the queue will be the first bus transaction (the last($)
        will be the final transaction

        Args:
            q:
        Raises:
        """
        raise Exception("UVMRegTransactionOrderPolicy::order pure virtual function")


#------------------------------------------------------------------------------
#
# Class: uvm_reg_map
#
# :Address map abstraction class
#
# This class represents an address map.
# An address map is a collection of registers and memories
# accessible via a specific physical interface.
# Address maps can be composed into higher-level address maps.
#
# Address maps are created using the <uvm_reg_block::create_map()>
# method.
#------------------------------------------------------------------------------


class UVMRegMap(UVMObject):


    def Xinit_address_mapX(self):
        bus_width = 0
        top_map = self.get_root_map()

        if self == top_map:
            top_map.m_regs_by_offset = {}
            top_map.m_regs_by_offset_wo = {}
            top_map.m_mems_by_offset = {}

        # TODO this piece of code
        for l in self.m_submaps:
            _map = l
            _map.Xinit_address_mapX()

        for rg_ in self.m_regs_info:
            rg = rg_  # uvm_reg
            self.m_regs_info[rg].is_initialized = True
            if not self.m_regs_info[rg].unmapped:
                rg_acc = rg.Xget_fields_accessX(self)
                addrs = []
                rg_offset = self.m_regs_info[rg].offset
                bus_width = self.get_physical_addresses(rg_offset, 0, rg.get_n_bytes(), addrs)

                for i in range(len(addrs)):
                    addr = addrs[i]

                    if addr in top_map.m_regs_by_offset:
                        rg2 = top_map.m_regs_by_offset[addr]
                        rg2_acc = rg2.Xget_fields_accessX(self)

                        # If the register at the same address is RO or WO
                        # and this register is WO or RO, this is OK
                        if (rg_acc == "RO" and rg2_acc == "WO"):
                            top_map.m_regs_by_offset[addr]    = rg
                            UVMRegReadOnlyCbs.add(rg)
                            top_map.m_regs_by_offset_wo[addr] = rg2
                            UVMRegWriteOnlyCbs.add(rg2)
                        elif (rg_acc == "WO" and rg2_acc == "RO"):
                            top_map.m_regs_by_offset_wo[addr] = rg
                            UVMRegWriteOnlyCbs.add(rg)
                            UVMRegReadOnlyCbs.add(rg2)
                        else:
                            a = ""
                            a = sv.sformatf("%0h",addr)
                            uvm_warning("RegModel", ("In map '" +
                                self.get_full_name() + "' register '" + rg.get_full_name()
                                + "' maps to same address as register '"
                                + top_map.m_regs_by_offset[addr].get_full_name()
                                + "': 'h" + a))
                    else:
                        top_map.m_regs_by_offset[addr] = rg

                    #foreach (top_map.m_mems_by_offset[range]):
                    for rr in top_map.m_mems_by_offset:
                        if (addr >= rr.min and addr <= rr.max):
                            a = sv.sformatf("%0h",addr)
                            b = sv.sformatf("[%0h:%0h]", rr.min, rr.max)
                            uvm_warning("RegModel", ("In map '",
                                    self.get_full_name() + "' register '" +
                                rg.get_full_name() + "' with address " + a +
                                "maps to same address as memory '" +
                                top_map.m_mems_by_offset[rr].get_full_name()
                                + "': " + b))
                self.m_regs_info[rg].addr = addrs

        # TODO complete this
        for mem_ in self.m_mems_info:
            mem = mem_
            if not self.m_mems_info[mem].unmapped:
                addrs = []
                addrs_max = []
                min1 = 0
                max1 = 0
                min2 = 0
                max2 = 0  # uvm_reg_addr_t
                stride = 0

                mem_offset = self.m_mems_info[mem].offset

                bus_width = self.get_physical_addresses(mem_offset,0,mem.get_n_bytes(),addrs)
                if addrs[0] < addrs[len(addrs)-1]:
                    min1 = addrs[0]
                else:
                    min1 = addrs[len(addrs)-1]
                min2 = addrs[0]

                mem_offset = self.m_mems_info[mem].offset
                self.get_physical_addresses(mem_offset, mem.get_size()-1, mem.get_n_bytes(), addrs_max)
                if addrs_max[0] > addrs_max[len(addrs_max)-1]:
                    max1 = addrs_max[0]
                else:
                    max1 = addrs_max[len(addrs_max)-1]
                max2 = addrs_max[0]
                # address interval between consecutive mem offsets
                stride = int((max2 - min2)/(mem.get_size() - 1))

                for reg_addr in top_map.m_regs_by_offset:
                    if (reg_addr >= min1 and reg_addr <= max1):
                        a = ""
                        a = sv.sformatf("%0h",reg_addr)
                        reg_name = top_map.m_regs_by_offset[reg_addr].get_full_name()
                        uvm_warning("RegModel", ("In map '" + self.get_full_name()
                            + "' memory '" + mem.get_full_name() + "' maps to same address as register '"
                            + reg_name + "': 'h" + a))

                for rr in top_map.m_mems_by_offset:
                    if (min1 <= rr.max and max1 >= rr.max or
                            min1 <= rr.min and max1 >= rr.min or
                            min1 >= rr.min and max1 <= rr.max):
                        a = sv.sformatf("[%0h:%0h]",min,max)
                        uvm_warning("RegModel", ("In map '" + self.get_full_name()
                            + "' memory '" + mem.get_full_name() + "' overlaps with address range of memory '"
                            + top_map.m_mems_by_offset[rr].get_full_name()
                            + "': 'h" + a))

                addr_range = UVMRegMapAddrRange(min1, max1, stride)
                top_map.m_mems_by_offset[addr_range] = mem
                self.m_mems_info[mem].addr  = addrs
                self.m_mems_info[mem].mem_range = addr_range

        # If the block has no registers or memories,
        # bus_width won't be set
        if bus_width == 0:
            bus_width = self.m_n_bytes
        self.m_system_n_bytes = bus_width
        #endfunction

    # static local uvm_reg_map   m_backdoor;
    m_backdoor = None  # uvm_reg_map

    @classmethod
    def backdoor(cls):
        """
           Function: backdoor
           Return the backdoor pseudo-map singleton

           This pseudo-map is used to specify or configure the backdoor
           instead of a real address map.

        Returns:
        """
        if UVMRegMap.m_backdoor is None:
            UVMRegMap.m_backdoor = UVMRegMap("Backdoor")
        return UVMRegMap.m_backdoor

    #   //----------------------
    #   // Group: Initialization
    #   //----------------------

    def __init__(self, name="uvm_reg_map"):
        """
           Function: new

           Create a new instance

        Args:
            name:
        """
        n = "default_map"
        if name != "":
            n = name
        UVMObject.__init__(self, n)
        self.m_auto_predict = 0
        self.m_check_on_read = True
        # info that is valid only if top-level map
        self.m_base_addr = 0  # uvm_reg_addr_t
        self.m_n_bytes = 0  # int
        self.m_endian = 0  # uvm_endianness_e
        self.m_byte_addressing = False
        self.m_sequence_wrapper = None  # uvm_object_wrapper
        self.m_adapter = None  # uvm_reg_adapter
        self.m_sequencer = None  # uvm_sequencer_base
        self.m_parent = None  # uvm_reg_block
        self.m_system_n_bytes = 0
        #
        self.m_parent_map = None  # uvm_reg_map
        self.m_parent_maps = {}  # uvm_reg_addr_t[uvm_reg_map] // value=offset of this map at parent level
        self.m_submaps = {}  # uvm_reg_addr_t[uvm_reg_map] value=offset of submap at this level
        self.m_submap_rights = {}  # string[uvm_reg_map] # value=rights of submap at this level

        self.m_regs_info = {}  # UVMRegMapInfo[uvm_reg]
        self.m_mems_info = {}  # UVMRegMapInfo[uvm_mem]
        self.m_regs_by_offset = {}  # uvm_reg[uvm_reg_addr_t]
        # Use only in addition to above if a RO and a WO
        # register share the same address.
        self.m_regs_by_offset_wo = {}  # uvm_reg[uvm_reg_addr_t]
        self.m_mems_by_offset = {}  # uvm_mem[UVMRegMapAddrRange]
        self.policy = None  # local uvm_reg_transaction_order_policy

    def configure(self, parent, base_addr, n_bytes, endian, byte_addressing=1):
        """
           Function: configure

           Instance-specific configuration

           Configures this map with the following properties.

           parent    - the block in which this map is created and applied

           base_addr - the base address for this map. All registers, memories,
                       and sub-blocks will be at offsets to this address

           n_bytes   - the byte-width of the bus on which this map is used

           endian    - the endian format. See `uvm_endianness_e` for possible
                       values

           byte_addressing - specifies whether the address increment is on a
                       per-byte basis. For example, consecutive memory locations
                       with `n_bytes`=4 (32-bit bus) are 4 apart: 0, 4, 8, and
                       so on. Default is TRUE.

        Args:
            parent:
            base_addr:
            n_bytes:
            endian:
            byte_addressing:
        """
        self.m_parent     = parent
        self.m_n_bytes    = n_bytes
        self.m_endian     = endian
        self.m_base_addr  = base_addr
        self.m_byte_addressing = byte_addressing

    def add_reg(self, rg, offset, rights="RW", unmapped=0, frontdoor=None):
        """
           Function: add_reg

           Add a register

           Add the specified register instance `rg` to this address map.

           The register is located at the specified address `offset` from
           this maps configured base address.

           The `rights` specify the register's accessibility via this map.
           Valid values are "RW", "RO", and "WO". Whether a register field
           can be read or written depends on both the field's configured access
           policy (see <uvm_reg_field::configure> and the register's rights in
           the map being used to access the field.

           The number of consecutive physical addresses occupied by the register
           depends on the width of the register and the number of bytes in the
           physical interface corresponding to this address map.

           If `unmapped` is TRUE, the register does not occupy any
           physical addresses and the base address is ignored.
           Unmapped registers require a user-defined `frontdoor` to be specified.

           A register may be added to multiple address maps
           if it is accessible from multiple physical interfaces.
           A register may only be added to an address map whose parent block
           is the same as the register's parent block.

        Args:
            rg:
            offset:
            rights:
            unmapped:
            frontdoor:
        """
        if rg in self.m_regs_info:
            uvm_error("RegModel", ("Register '" + rg.get_name()
                      + "' has already been added to map '" + self.get_name() + "'"))
            return

        if rg.get_parent() != self.get_parent():
            uvm_error("RegModel", ("Register '" + rg.get_full_name()
                + "' may not be added to address map '"
                + self.get_full_name() + "' : they are not in the same block"))
            return

        rg.add_map(self)
        info = UVMRegMapInfo()
        info.offset   = offset
        info.rights   = rights
        info.unmapped = unmapped
        info.frontdoor = frontdoor
        self.m_regs_info[rg] = info
    #endfunction


    def add_mem(self, mem, offset, rights="RW", unmapped=0, frontdoor=None):
        """
           Function: add_mem

           Add a memory

           Add the specified memory instance to this address map.
           The memory is located at the specified base address and has the
           specified access rights ("RW", "RO" or "WO").
           The number of consecutive physical addresses occupied by the memory
           depends on the width and size of the memory and the number of bytes in the
           physical interface corresponding to this address map.

           If `unmapped` is TRUE, the memory does not occupy any
           physical addresses and the base address is ignored.
           Unmapped memories require a user-defined `frontdoor` to be specified.

           A memory may be added to multiple address maps
           if it is accessible from multiple physical interfaces.
           A memory may only be added to an address map whose parent block
           is the same as the memory's parent block.

          extern virtual function void add_mem (uvm_mem        mem,
                                                uvm_reg_addr_t offset,
                                                string         rights = "RW",
                                                bit            unmapped=0,
                                                uvm_reg_frontdoor frontdoor=None)
        Args:
            mem:
            offset:
            rights:
            unmapped:
            frontdoor:
        """
        if mem in self.m_mems_info:
            uvm_error("RegModel", "Memory '" + mem.get_name() +
                      "' has already been added to map '" + self.get_name() + "'")
            return

        if mem.get_parent() != self.get_parent():
            uvm_error("RegModel", "Memory '" + mem.get_full_name() + "' may not be added to address map '"
                + self.get_full_name() + "' : they are not in the same block")
            return

        mem.add_map(self)

        info = UVMRegMapInfo()
        info.offset   = offset
        info.rights   = rights
        info.unmapped = unmapped
        info.frontdoor = frontdoor
        self.m_mems_info[mem] = info
        #endfunction: add_mem

    def add_submap(self, child_map, offset):
        """
           Function: add_submap

           Add an address map

           Add the specified address map instance to this address map.
           The address map is located at the specified base address.
           The number of consecutive physical addresses occupied by the submap
           depends on the number of bytes in the physical interface
           that corresponds to the submap,
           the number of addresses used in the submap and
           the number of bytes in the
           physical interface corresponding to this address map.

           An address map may be added to multiple address maps
           if it is accessible from multiple physical interfaces.
           An address map may only be added to an address map
           in the grand-parent block of the address submap.

          extern virtual function void add_submap (uvm_reg_map    child_map,
                                                   uvm_reg_addr_t offset)
        Args:
            child_map:
            offset:
        """
        parent_map = None  # #   uvm_reg_map

        if child_map is None:
            uvm_error("RegModel", "Attempting to add None map to map '" +
                    self.get_full_name() + "'")
            return

        parent_map = child_map.get_parent_map()

        # Cannot have more than one parent (currently)
        if parent_map is not None:
            uvm_error("RegModel", ("Map '" + child_map.get_full_name() +
                      "' is already a child of map '" +
                      parent_map.get_full_name() +
                      "'. Cannot also be a child of map '" +
                      self.get_full_name() + "'"))
            return

        # begin : parent_block_check
        child_blk = child_map.get_parent()
        if child_blk is None:
            uvm_error("RegModel", "Cannot add submap '" +child_map.get_full_name()
                  + "' because it does not have a parent block")
            return

        while ((child_blk is not None) and (child_blk.get_parent() != self.get_parent())):
            child_blk = child_blk.get_parent()

        if child_blk is None:
            uvm_error("RegModel", ("Submap '" + child_map.get_full_name()
                + "' may not be added to this address map, '"
                + self.get_full_name(),"', as the submap's parent block, '"
                + child_blk.get_full_name()
                + "', is neither this map's parent block nor a descendent of this map's parent block, '"
                + self.m_parent.get_full_name() + "'"))
            return


        if self.m_n_bytes > child_map.get_n_bytes(UVM_NO_HIER):
            uvm_warning("RegModel",
               sv.sformatf("Adding %0d-byte submap '%s' to %0d-byte parent map '%s'",
                         child_map.get_n_bytes(UVM_NO_HIER), child_map.get_full_name(),
                         self.m_n_bytes, self.get_full_name()))


        child_map.add_parent_map(self, offset)
        self.set_submap_offset(child_map, offset)
        #endfunction: add_submap


    def set_sequencer(self, sequencer, adapter=None):
        """
           Function: set_sequencer

           Set the sequencer and adapter associated with this map. This method
           `must` be called before starting any sequences based on uvm_reg_sequence.

          extern virtual function void set_sequencer (uvm_sequencer_base sequencer,
                                                      uvm_reg_adapter    adapter=None)
        Args:
            sequencer:
            adapter:
        """
        if sequencer is None:
            uvm_error("REG_None_SQR", "None reference specified for bus sequencer")
            return

        if adapter is None:
            uvm_info("REG_NO_ADAPT", ("Adapter not specified for map '" +
                self.get_full_name() +
                "'. Accesses via this map will send abstract 'uvm_reg_item' items to sequencer '"
                + sequencer.get_full_name() + "'"), UVM_MEDIUM)

        self.m_sequencer = sequencer
        self.m_adapter = adapter
        #endfunction

    def set_submap_offset(self, submap, offset):
        """

           Function: set_submap_offset

           Set the offset of the given `submap` to `offset`.

          extern virtual function void set_submap_offset (uvm_reg_map submap,
                                                          uvm_reg_addr_t offset)
        Args:
            submap:
            offset:
        """
        if (submap is None):
            uvm_error("REG/None","set_submap_offset: submap handle is None")
            return

        self.m_submaps[submap] = offset
        if self.m_parent.is_locked():
            root_map = self.get_root_map()
            root_map.Xinit_address_mapX()


    def get_submap_offset(self, submap):
        """
           Function: get_submap_offset

           Return the offset of the given `submap`.

          extern virtual function uvm_reg_addr_t get_submap_offset (uvm_reg_map submap)
        Args:
            submap:
        Returns:
        """
        if submap is None:
            uvm_error("REG/None","set_submap_offset: submap handle is None")
            return -1

        if submap not in self.m_submaps:
            uvm_error("RegModel", "Map '" + submap.get_full_name()
                  + "' is not a submap of '" + self.get_full_name() + "'")
            return -1
        return self.m_submaps[submap]


    #   // Function: set_base_addr
    #   //
    #   // Set the base address of this map.
    #   extern virtual function void   set_base_addr (uvm_reg_addr_t  offset)


    def reset(self, kind="SOFT"):
        """
           Function: reset

           Reset the mirror for all registers in this address map.

           Sets the mirror value of all registers in this address map
           and all of its submaps
           to the reset value corresponding to the specified reset event.
           See <uvm_reg_field::reset()> for more details.
           Does not actually set the value of the registers in the design,
           only the values mirrored in their corresponding mirror.

           Note that, unlike the other reset() method, the default
           reset event for this method is "SOFT".

          extern virtual function void reset(string kind = "SOFT")
        Args:
            kind:
        """
        regs = []
        self.get_registers(regs)
        for reg in regs:
            reg.reset(kind)


    def add_parent_map(self, parent_map, offset):
        """
          /*local*/ extern virtual function void add_parent_map(uvm_reg_map  parent_map,
                                                                uvm_reg_addr_t offset)
        Args:
            parent_map:
            offset:
        """
        if (parent_map is None):
            uvm_error("RegModel",
               "Attempting to add None parent map to map '" +
               self.get_full_name() + "'")
            return

        if self.m_parent_map is not None:
            uvm_error("RegModel",
               sv.sformatf("Map \"%s\" already a submap of map \"%s\" at offset 'h%h",
                         self.get_full_name(), self.m_parent_map.get_full_name(),
                         self.m_parent_map.get_submap_offset(self)))
            return


        self.m_parent_map = parent_map
        self.m_parent_maps[parent_map] = offset  # prep for multiple parents
        parent_map.m_submaps[self] = offset
        #
        #endfunction: add_parent_map

    #   /*local*/ extern virtual function void Xverify_map_configX()


    def m_set_reg_offset(self, rg, offset, unmapped):
        """
        m_set_reg_offset
        Args:
            rg:
            offset:
            unmapped:
        """
        if not rg in self.m_regs_info:
            uvm_error("RegModel",
              ("Cannot modify offset of register '" + rg.get_full_name()
              + "' in address map '" + self.get_full_name()
              + "' : register not mapped in that address map"))
            return

        info    = self.m_regs_info[rg]
        blk     = self.get_parent()
        top_map = self.get_root_map()
        addrs = []

        if blk is None:
            uvm_fatal("RegModel", "Addr map requires parent reg_block")

        # if block is not locked, Xinit_address_mapX will resolve map when block is locked
        if blk.is_locked():

            # remove any existing cached addresses
            if info.unmapped is False:
                for addr in info.addr:
                    if addr not in top_map.m_regs_by_offset_wo:
                        del top_map.m_regs_by_offset[addr]
                    else:
                        if top_map.m_regs_by_offset[addr] == rg:
                            top_map.m_regs_by_offset[addr] = top_map.m_regs_by_offset_wo[addr]
                            #TODO UVMRegReadOnlyCbs.remove(rg)
                            #TODO UVMRegWriteOnlyCbs.remove(top_map.m_regs_by_offset[info.addr[i]])
                        else:
                            pass
                            #TODO UVMRegWriteOnlyCbs.remove(rg)
                            #TODO UVMRegReadOnlyCbs.remove(top_map.m_regs_by_offset[info.addr[i]])
                        del top_map.m_regs_by_offset_wo[addr]

            # if we are remapping...
            if not unmapped:
                rg_acc = rg.Xget_fields_accessX(self)
                # get new addresses
                self.get_physical_addresses(offset,0,rg.get_n_bytes(),addrs)

                # make sure they do not conflict with others
                for adr in addrs:
                    addr = adr

                    if addr in top_map.m_regs_by_offset:
                        rg2 = top_map.m_regs_by_offset[addr]
                        rg2_acc = rg2.Xget_fields_accessX(self)

                        # If the register at the same address is RO or WO
                        # and this register is WO or RO, this is OK
                        if (rg_acc == "RO" and rg2_acc == "WO"):
                           top_map.m_regs_by_offset[addr]    = rg
                           # TODO UVMRegReadOnlyCbs.add(rg)
                           top_map.m_regs_by_offset_wo[addr] = rg2
                           # TODO UVMRegWriteOnlyCbs.add(rg2)
                        elif (rg_acc == "WO" and rg2_acc == "RO"):
                           top_map.m_regs_by_offset_wo[addr] = rg
                           # TODO UVMRegWriteOnlyCbs.add(rg)
                           # TODO UVMRegReadOnlyCbs.add(rg2)
                        else:
                            a = "{}".format(addr)
                            uvm_warning("RegModel", ("In map '" + self.get_full_name()
                                + "' register '"+ rg.get_full_name()
                                + "' maps to same address as register '"
                                + top_map.m_regs_by_offset[addr].get_full_name()
                                + "': 'h" + a))
                    else:
                        top_map.m_regs_by_offset[addr] = rg

                    for range in top_map.m_mems_by_offset.keys():
                        if adr >= range.min and adr <= range.max:
                            a = "{}".format(adr)
                            uvm_warning("RegModel", ("In map '" + self.get_full_name()
                                + "' register '" +
                                rg.get_full_name() + "' overlaps with address range of memory '"
                                + top_map.m_mems_by_offset[range].get_full_name()
                                + "': 'h" +a))

                info.addr = addrs # cache it

        if unmapped is True:
            info.offset   = -1
            info.unmapped = 1
        else:
          info.offset   = offset
          info.unmapped = 0
    #endfunction

    #   /*local*/ extern virtual function void m_set_mem_offset(uvm_mem mem,
    #                                                           uvm_reg_addr_t offset,
    #                                                           bit unmapped)
    #


    #   //---------------------
    #   // Group: Introspection
    #   //---------------------
    #
    #   // Function: get_name
    #   //
    #   // Get the simple name
    #   //
    #   // Return the simple object name of this address map.
    #   //

    def get_full_name(self):
        """
           Function: get_full_name

           Get the hierarchical name

           Return the hierarchal name of this address map.
           The base of the hierarchical name is the root block.

          extern virtual function string get_full_name()
        Returns:
        """
        get_full_name = self.get_name()
        if self.m_parent is None:
            return get_full_name
        return self.m_parent.get_full_name() + "." + get_full_name

    def get_root_map(self):
        """
        Get the externally-visible address map

        Get the top-most address map where this address map is instantiated.
        It corresponds to the externally-visible address map that can
        be accessed by the verification environment.

        Returns:
            UVMRegMap: Root map for this address map.
        """
        if self.m_parent_map is None:
            return self
        else:
            return self.m_parent_map.get_root_map()

    def get_parent(self):
        """
           Function: get_parent

           Get the parent block

           Return the block that is the parent of this address map.

        Returns:
        """
        return self.m_parent

    def get_parent_map(self):
        """
           Function: get_parent_map
           Get the higher-level address map

           Return the address map in which this address map is mapped.
           returns `None` if this is a top-level address map.

          extern virtual function uvm_reg_map           get_parent_map()
        Returns:
        """
        return self.m_parent_map


    def get_base_addr(self, hier=UVM_HIER):
        """
           Function: get_base_addr

           Get the base offset address for this map. If this map is the
           root map, the base address is that set with the `base_addr` argument
           to <uvm_reg_block::create_map()>. If this map is a submap of a higher-level map,
           the base address is offset given this submap by the parent map.
           See `set_submap_offset`.

          extern virtual function uvm_reg_addr_t get_base_addr (uvm_hier_e hier=UVM_HIER)
        Args:
            hier:
            UVM_HIER:
        Returns:
        """
        # child = self
        if (hier == UVM_NO_HIER or self.m_parent_map is None):
            return self.m_base_addr
        get_base_addr = self.m_parent_map.get_submap_offset(self)
        get_base_addr += self.m_parent_map.get_base_addr(UVM_HIER)
        return get_base_addr


    def get_n_bytes(self, hier=UVM_HIER):
        """
           Function: get_n_bytes

           Get the width in bytes of the bus associated with this map. If `hier`
           is `UVM_HIER`, then gets the effective bus width relative to the system
           level. The effective bus width is the narrowest bus width from this
           map to the top-level root map. Each bus access will be limited to this
           bus width.

          extern virtual function int unsigned get_n_bytes (uvm_hier_e hier=UVM_HIER)
        Args:
            hier:
            UVM_HIER:
        Returns:
        """
        if hier == UVM_NO_HIER:
            return self.m_n_bytes
        return self.m_system_n_bytes


    def get_addr_unit_bytes(self):
        """
           Function: get_addr_unit_bytes

           Get the number of bytes in the smallest addressable unit in the map.
           Returns 1 if the address map was configured using byte-level addressing.
           Returns <get_n_bytes()> otherwise.

          extern virtual function int unsigned get_addr_unit_bytes()
        Returns:
        """
        if (self.m_byte_addressing):
            return 1
        else:
            return self.m_n_bytes


    def get_endian(self, hier=UVM_HIER):
        """
           Function: get_base_addr

           Gets the endianness of the bus associated with this map. If `hier` is
           set to `UVM_HIER`, gets the system-level endianness.

          extern virtual function uvm_endianness_e get_endian (uvm_hier_e hier=UVM_HIER)
        Args:
            hier:
            UVM_HIER:
        Returns:
        """
        if (hier == UVM_NO_HIER or self.m_parent_map is None):
            return self.m_endian
        return self.m_parent_map.get_endian(hier)


    def get_sequencer(self, hier=UVM_HIER):
        """
           Function: get_sequencer

           Gets the sequencer for the bus associated with this map. If `hier` is
           set to `UVM_HIER`, gets the sequencer for the bus at the system-level.
           See `set_sequencer`.

          extern virtual function uvm_sequencer_base get_sequencer (uvm_hier_e hier=UVM_HIER)
        Args:
            hier:
            UVM_HIER:
        Returns:
        """
        if (hier == UVM_NO_HIER or self.m_parent_map is None):
            return self.m_sequencer
        return self.m_parent_map.get_sequencer(hier)

    def get_adapter(self, hier=UVM_HIER):
        """
           Function: get_adapter

           Gets the bus adapter for the bus associated with this map. If `hier` is
           set to `UVM_HIER`, gets the adapter for the bus used at the system-level.
           See `set_sequencer`.

          extern virtual function uvm_reg_adapter get_adapter (uvm_hier_e hier=UVM_HIER)
        Args:
            hier:
            UVM_HIER:
        Returns:
        """
        if (hier == UVM_NO_HIER or self.m_parent_map is None):
            return self.m_adapter
        return self.m_parent_map.get_adapter(hier)
        #endfunction

    def get_submaps(self, maps, hier=UVM_HIER):
        """
           Function: get_submaps

           Get the address sub-maps

           Get the address maps instantiated in this address map.
           If `hier` is `UVM_HIER`, recursively includes the address maps,
           in the sub-maps.

          extern virtual function void  get_submaps (ref uvm_reg_map maps[$],
                                                     input uvm_hier_e hier=UVM_HIER)
        Args:
            maps:
            hier:
            UVM_HIER:
        """
        for submap in self.m_submaps:
            maps.append(submap)

        if (hier == UVM_HIER):
            for sm in self.m_submaps:
                submap = sm
                submap.get_submaps(maps)
        #endfunction

    def get_registers(self, regs, hier=UVM_HIER):
        """
           Function: get_registers

           Get the registers

           Get the registers instantiated in this address map.
           If `hier` is `UVM_HIER`, recursively includes the registers
           in the sub-maps.

          extern virtual function void  get_registers (ref uvm_reg regs[$],
                                                       input uvm_hier_e hier=UVM_HIER)
        Args:
            regs:
            hier:
            UVM_HIER:
        """
        for rg in self.m_regs_info:
            regs.append(rg)

        if hier == UVM_HIER:
            for sm in self.m_submaps:
                submap = sm
                submap.get_registers(regs)
        #endfunction

    #
    #   // Function: get_fields
    #   //
    #   // Get the fields
    #   //
    #   // Get the fields in the registers instantiated in this address map.
    #   // If ~hier~ is ~UVM_HIER~, recursively includes the fields of the registers
    #   // in the sub-maps.
    #   //
    #   extern virtual function void  get_fields (ref uvm_reg_field fields[$],
    #                                             input uvm_hier_e hier=UVM_HIER)
    #
    #
    #   // Function: get_memories
    #   //
    #   // Get the memories
    #   //
    #   // Get the memories instantiated in this address map.
    #   // If ~hier~ is ~UVM_HIER~, recursively includes the memories
    #   // in the sub-maps.
    #   //
    #   extern virtual function void  get_memories (ref uvm_mem mems[$],
    #                                               input uvm_hier_e hier=UVM_HIER)
    #
    #
    #   // Function: get_virtual_registers
    #   //
    #   // Get the virtual registers
    #   //
    #   // Get the virtual registers instantiated in this address map.
    #   // If ~hier~ is ~UVM_HIER~, recursively includes the virtual registers
    #   // in the sub-maps.
    #   //
    #   extern virtual function void  get_virtual_registers (ref uvm_vreg regs[$],
    #                                                        input uvm_hier_e hier=UVM_HIER)
    #
    #
    #   // Function: get_virtual_fields
    #   //
    #   // Get the virtual fields
    #   //
    #   // Get the virtual fields from the virtual registers instantiated
    #   // in this address map.
    #   // If ~hier~ is ~UVM_HIER~, recursively includes the virtual fields
    #   // in the virtual registers in the sub-maps.
    #   //
    #   extern virtual function void  get_virtual_fields (ref uvm_vreg_field fields[$],
    #                                                     input uvm_hier_e hier=UVM_HIER)
    #


    def get_reg_map_info(self, rg, error=1):
        """
        get_reg_map_info
        Args:
            rg:
            error:
        Returns:
        """
        result = None
        if rg not in self.m_regs_info:
            if error:
                uvm_error("REG_NO_MAP", ("Register '" + rg.get_name()
                    + "' not in map '" + self.get_name() + "'"))
            return None
        result = self.m_regs_info[rg]
        if not result.is_initialized:
            uvm_fatal("RegModel", ("map '" + self.get_name() +
                "' does not seem to be initialized correctly, "
                + "check that the top register model is locked()"))
        return result

    def get_mem_map_info(self, mem, error=1):
        """
          extern virtual function UVMRegMapInfo get_mem_map_info(uvm_mem mem, bit error=1)
        Args:
            mem:
            error:
        Returns:
        """
        if mem not in self.m_mems_info:
            if error:
                uvm_error("REG_NO_MAP", "Memory '" + mem.get_name() + "' not in map '"
                        + self.get_name() + "'")
            return None
        return self.m_mems_info[mem]


    #   extern virtual function int unsigned get_size()


    def get_physical_addresses(self, base_addr, mem_offset, n_bytes, addr):
        """

           Function: get_physical_addresses

           Translate a local address into external addresses

           Identify the sequence of addresses that must be accessed physically
           to access the specified number of bytes at the specified address
           within this address map.
           Returns the number of bytes of valid data in each access.

           Returns in `addr` a list of address in little endian order,
           with the granularity of the top-level address map.

           A register is specified using a base address with `mem_offset` as 0.
           A location within a memory is specified using the base address
           of the memory and the index of the location within that memory.


          extern virtual function int get_physical_addresses(uvm_reg_addr_t        base_addr,
                                                             uvm_reg_addr_t        mem_offset,
                                                             int unsigned          n_bytes,
                                                             ref uvm_reg_addr_t    addr[])
        Args:
            base_addr:
            mem_offset:
            n_bytes:
            addr:
        Returns:
        """
        bus_width = self.get_n_bytes(UVM_NO_HIER)
        up_map = None  # uvm_reg_map
        local_addr = []
        multiplier = 1
        if self.m_byte_addressing:
            multiplier = bus_width
        #addr = new [0]

        if (n_bytes <= 0):
           uvm_fatal("RegModel", sv.sformatf(
               "Cannot access %0d bytes. Must be greater than 0", n_bytes))
           return 0

        # First, identify the addresses within the block/system
        if n_bytes <= bus_width:
            local_addr.append(base_addr + (mem_offset * multiplier))
        else:
            n = 0
            n = int(((n_bytes-1) / bus_width) + 1)
            local_addr = [0] * n
            base_addr = base_addr + mem_offset * (n * multiplier)
            get_end = self.get_endian(UVM_NO_HIER)

            if get_end == UVM_LITTLE_ENDIAN:
                for i in range(len(local_addr)):
                    local_addr[i] = base_addr + (i * multiplier)
            elif get_end == UVM_BIG_ENDIAN:
                for i in range(len(local_addr)):
                    n -= 1
                    local_addr[i] = base_addr + (n * multiplier)
            elif get_end == UVM_LITTLE_FIFO:
                for i in range(len(local_addr)):
                    local_addr[i] = base_addr
            elif get_end == UVM_BIG_FIFO:
                for i in range(len(local_addr)):
                    local_addr[i] = base_addr
            else:
                uvm_error("RegModel",
                        ("Map has no specified endianness. ",
                        sv.sformatf("Cannot access %0d bytes register via its %0d byte %s interface",
                        n_bytes, bus_width, self.get_full_name())))

        up_map = self.get_parent_map()

        # Then translate these addresses in the parent's space
        if up_map is None:
            # This is the top-most system/block!
            addr.clear()
            addr.extend(local_addr)
            for i in range(len(addr)):
                addr[i] += self.m_base_addr
        else:
            sys_addr = []
            base_addr = 0
            w = 0
            k = 0

            # Scale the consecutive local address in the system's granularity
            if (bus_width < up_map.get_n_bytes(UVM_NO_HIER)):
                k = 1
            else:
                k = int((bus_width-1) / up_map.get_n_bytes(UVM_NO_HIER)) + 1

            base_addr = up_map.get_submap_offset(self)
            for i in range(len(local_addr)):
                n = len(addr)
                w = up_map.get_physical_addresses(base_addr + local_addr[i] * k,
                        0, bus_width, sys_addr)

                #addr = new [n + sys_addr.size()] (addr)
                for j in range(len(sys_addr)):
                    addr.append(sys_addr[j])
            # The width of each access is the minimum of this block or the system's width
            if (w < bus_width):
                bus_width = w

        return bus_width
        #
        #endfunction: get_physical_addresses


    def get_reg_by_offset(self, offset, read=True):
        """


           Function: get_reg_by_offset

           Get register mapped at offset

           Identify the register located at the specified offset within
           this address map for the specified type of access.
           Returns `None` if no such register is found.

           The model must be locked using <uvm_reg_block::lock_model()>
           to enable this functionality.

          extern virtual function uvm_reg get_reg_by_offset(uvm_reg_addr_t offset,
                                                            bit            read = 1)
        Args:
            offset:
            read:
        Returns:
        """
        if not(self.m_parent.is_locked()):
            uvm_error("RegModel", sv.sformatf(
                "Cannot get register by offset: Block %s is not locked.",
                self.m_parent.get_full_name()))
            return None

        if (not read and offset in self.m_regs_by_offset_wo):
            return self.m_regs_by_offset_wo[offset]

        if offset in self.m_regs_by_offset:
            return self.m_regs_by_offset[offset]

        return None


    #   //
    #   // Function: get_mem_by_offset
    #   // Get memory mapped at offset
    #   //
    #   // Identify the memory located at the specified offset within
    #   // this address map. The offset may refer to any memory location
    #   // in that memory.
    #   // Returns ~None~ if no such memory is found.
    #   //
    #   // The model must be locked using <uvm_reg_block::lock_model()>
    #   // to enable this functionality.
    #   //
    #   extern virtual function uvm_mem    get_mem_by_offset(uvm_reg_addr_t offset)


    #   //------------------
    #   // Group: Bus Access
    #   //------------------


    def set_auto_predict(self, on=True):
        """
        Sets the auto-predict mode for his map.

        When `on` is `TRUE`,
        the register model will automatically update its mirror
        (what it thinks should be in the DUT) immediately after
        any bus read or write operation via this map. Before a `UVMReg.write`
        or `UVMReg.read` operation returns, the register's `UVMReg.predict`
        method is called to update the mirrored value in the register.

        When `on` is `FALSE`, bus reads and writes via this map do not
        automatically update the mirror. For real-time updates to the mirror
        in this mode, you connect a `uvm_reg_predictor` instance to the bus
        monitor. The predictor takes observed bus transactions from the
        bus monitor, looks up the associated `UVMReg` register given
        the address, then calls that register's `UVMReg.predict` method.
        While more complex, this mode will capture all register read/write
        activity, including that not directly descendant from calls to
        `UVMReg.write` and `UVMReg.read`.

        By default, auto-prediction is turned off.

        Args:
            on:
        """
        self.m_auto_predict = on

    def get_auto_predict(self):
        """
        Gets the auto-predict mode setting for this map.

        Returns:
            bool: True if auto-predict enabled, False otherwise.
        """
        return self.m_auto_predict


    def set_check_on_read(self, on=1):
        """
        Sets the check-on-read mode for his map
        and all of its submaps.

        When `on` is `TRUE`,
        the register model will automatically check any value read back from
        a register or field against the current value in its mirror
        and report any discrepancy.
        This effectively combines the functionality of the
        <uvm_reg::read()> and ~uvm_reg::mirror(UVM_CHECK)~ method.
        This mode is useful when the register model is used passively.

        When `on` is `FALSE`, no check is made against the mirrored value.

        At the end of the read operation, the mirror value is updated based
        on the value that was read regardless of this mode setting.

        By default, auto-prediction is turned off.

        Args:
            on:
        """
        self.m_check_on_read = on
        for submap in self.m_submaps:
            submap.set_check_on_read(on)


    def get_check_on_read(self):
        """
        Gets the check-on-read mode setting for this map.

        Returns:
            bool: True if checking on read enabled, False otherwise
        """
        return self.m_check_on_read


    async def do_bus_write(self, rw, sequencer, adapter):
        """
        Perform a bus write operation.

        Args:
            rw (UVMRegItem):
            sequencer (UVMSequencer):
            adapter (UVMRegAdapter):
        """
        addrs = []
        system_map = self.get_root_map()
        bus_width = self.get_n_bytes()
        byte_en = sv.get_vector_of_ones(bus_width)
        map_info = None
        n_bits = 0
        lsb = 0
        skip = 0
        curr_byte = 0
        n_access_extra = 0
        n_access = 0
        n_bits_init = 0
        accesses = []

        [map_info, n_bits_init, lsb, skip] = self.Xget_bus_infoX(rw, map_info, n_bits_init, lsb, skip)
        #addrs = map_info.addr
        addrs = map_info.addr.copy()

        # if a memory, adjust addresses based on offset
        if rw.element_kind == UVM_MEM:
            for i in range(len(addrs)):
                addrs[i] = addrs[i] + map_info.mem_range.stride * rw.offset

        for val_idx in range(len(rw.value)):
            value = rw.value[val_idx]

            # /* calculate byte_enables */
            if rw.element_kind == UVM_FIELD:
                temp_be = 0
                idx = 0
                n_access_extra = lsb % (bus_width*8)
                n_access = n_access_extra + n_bits_init
                temp_be = n_access_extra
                value = value << n_access_extra
                while temp_be >= 8:
                    byte_en = sv.clear_bit(byte_en, idx)
                    idx += 1
                    temp_be -= 8

                temp_be += n_bits_init
                while temp_be > 0:
                    byte_en = sv.set_bit(byte_en, idx)
                    # byte_en[idx] = 1
                    idx += 1
                    temp_be -= 8
                byte_en &= (1 << idx)-1
                for i in range(skip):
                    addrs.pop_front()
                while len(addrs) > (int(n_bits_init/(bus_width*8)) + 1):
                    addrs.pop_back()
            curr_byte = 0
            n_bits = n_bits_init

            accesses = []
            for i in range(len(addrs)):
                rw_access = UVMRegBusOp()
                data = (value >> (curr_byte*8)) & ((1 << (bus_width * 8))-1)

                uvm_info(self.get_type_name(),
                   sv.sformatf("Writing 0x%0h at 0x%0h via map %s...",
                        data, addrs[i], rw.map.get_full_name()), UVM_VERB_MEM_MAP)

                if rw.element_kind == UVM_FIELD:
                    for z in range(bus_width):
                        bit_val = sv.get_bit(byte_en, curr_byte + z)
                        # rw_access.byte_en[z] = byte_en[curr_byte+z]
                        rw_access.byte_en = sv.set_bit(rw_access.byte_en, z, bit_val)

                rw_access.kind    = rw.kind
                rw_access.addr    = addrs[i]
                rw_access.data    = data
                #rw_access.n_bits  = (n_bits > bus_width*8) ? bus_width*8 : n_bits
                rw_access.n_bits = n_bits
                if (n_bits > bus_width*8):
                    rw_access.n_bits = bus_width*8
                rw_access.byte_en = byte_en

                accesses.append(rw_access)
                curr_byte += bus_width
                n_bits -= bus_width * 8

            # if set utilizy the order policy
            if (self.policy is not None):
                self.policy.order(accesses)

            # perform accesses
            # foreach(accesses[i]):
            for i in range(len(accesses)):
                rw_access = accesses[i]  # uvm_reg_bus_op
                bus_req = None  # uvm_sequence_item
                adapter.m_set_item(rw)
                bus_req = adapter.reg2bus(rw_access)
                adapter.m_set_item(None)

                if bus_req is None:
                    uvm_fatal("RegMem",
                        "adapter [" + adapter.get_name() + "] didnt return a bus transaction")

                bus_req.set_sequencer(sequencer)
                await rw.parent.start_item(bus_req, rw.prior)

                if (rw.parent is not None and i == 0):
                    rw.parent.mid_do(rw)

                await rw.parent.finish_item(bus_req)
                await bus_req.end_event.wait_on()

                if adapter.provides_responses:
                    bus_rsp = None  # uvm_sequence_item
                    op = None  # uvm_access_e
                    # TODO: need to test for right trans type, if not put back in q
                    await rw.parent.get_base_response(bus_rsp)
                    rw_access = adapter.bus2reg(bus_rsp, rw_access)
                else:
                    rw_access = adapter.bus2reg(bus_req, rw_access)
                    if rw_access is None:
                        uvm_error("ADAPTER_BUS2REG_NONE", sv.sformatf("Adapter %s"
                            + " returned None for RW item %s",
                            adapter.get_name(),
                            bus_req.convert2string())
                        )

                if (rw.parent is not None and i == len(addrs)-1):
                    rw.parent.post_do(rw)

                rw.status = rw_access.status

                uvm_info(self.get_type_name(),
                   sv.sformatf("Wrote 0x%0h at 0x%0h via map %s: %s...",
                      rw_access.data, addrs[i], rw.map.get_full_name(), rw.status), UVM_VERB_MEM_MAP)

                if rw.status == UVM_NOT_OK:
                    break

            for i in range(len(addrs)):
                addrs[i] = addrs[i] + map_info.mem_range.stride

        #end: foreach_value
        #
        #endtask: do_bus_write



    async def do_bus_read(self, rw, sequencer, adapter):
        """
        Perform a bus read operation.

        Args:
            rw (UVMRegItem):
            sequencer (UVMSequencer):
            adapter (UVMRegAdapter):
        """
        addrs = []  # uvm_reg_addr_t[$]
        system_map = self.get_root_map()
        bus_width = self.get_n_bytes()
        byte_en = sv.get_vector_of_ones(bus_width)
        map_info = None
        size = 0
        n_bits = 0
        skip = 0
        lsb = 0
        curr_byte = 0
        n_access_extra = 0
        n_access = 0
        n_bits_init = 0
        accesses = []  # uvm_reg_bus_op[$]

        # print("do_bus_read Given rw was " + rw.convert2string())
        [map_info, n_bits_init, lsb, skip] = self.Xget_bus_infoX(rw, map_info, n_bits_init, lsb, skip)
        # Need a copy as this will be modified later
        addrs = map_info.addr.copy()

        # if a memory, adjust addresses based on offset
        if (rw.element_kind == UVM_MEM):
            for i in range(len(addrs)):
                addrs[i] = addrs[i] + map_info.mem_range.stride * rw.offset

        for val_idx in range(len(rw.value)):
            rw.value[val_idx] = 0

            # /* calculate byte_enables */
            if (rw.element_kind == UVM_FIELD):
                temp_be = 0
                ii = 0
                n_access_extra = lsb % (bus_width*8)
                n_access = n_access_extra + n_bits_init
                temp_be = n_access_extra
                while (temp_be >= 8):
                    # byte_en[ii] = 0
                    byte_en = sv.clear_bit(byte_en, ii)
                    ii += 1
                    temp_be -= 8

                temp_be += n_bits_init
                while(temp_be > 0):
                    # byte_en[ii] = 1
                    byte_en = sv.set_bit(byte_en, ii)
                    ii += 1
                    temp_be -= 8
                byte_en &= (1 << ii) - 1
                for i in range(skip):
                    addrs.pop_front()
                while addrs.size() > (int(n_bits_init/(bus_width*8)) + 1):
                    addrs.pop_back()
            #end
            curr_byte = 0
            n_bits = n_bits_init

            accesses = []
            for i in range(len(addrs)):
                rw_access = UVMRegBusOp()

                uvm_info(self.get_type_name(),
                   sv.sformatf("Reading address 'h%0h via map \"%s\"...",
                             addrs[i], self.get_full_name()), UVM_VERB_MEM_MAP)

                if (rw.element_kind == UVM_FIELD):
                    #for (int z=0;z<bus_width;z++)
                    for z in range(bus_width):
                        # rw_access.byte_en[z] = byte_en[curr_byte+z]
                        bit_val = sv.get_bit(byte_en, curr_byte + z)
                        rw_access.byte_en = sv.set_bit(rw_access.byte_en, z, bit_val)

                rw_access.kind = rw.kind
                rw_access.addr = addrs[i]
                rw_access.data = curr_byte
                rw_access.byte_en = byte_en
                rw_access.n_bits = n_bits
                if (n_bits > bus_width*8):
                    rw_access.n_bits = bus_width*8

                accesses.append(rw_access)

                curr_byte += bus_width
                n_bits -= bus_width * 8

            # if set utilize the order policy
            if(self.policy is not None):
                self.policy.order(accesses)

            # perform accesses
            for i in range(len(accesses)):
                rw_access = accesses[i]  # uvm_reg_bus_op
                bus_req = None  # uvm_sequence_item
                data = 0  # uvm_reg_data_logic_t
                curr_byte_ = 0

                curr_byte_ = rw_access.data
                rw_access.data = 0
                adapter.m_set_item(rw)
                bus_req = adapter.reg2bus(rw_access)
                adapter.m_set_item(None)
                if bus_req is None:
                    uvm_fatal("RegMem","adapter [" + adapter.get_name() + "] didnt return a bus transaction")

                bus_req.set_sequencer(sequencer)
                await rw.parent.start_item(bus_req, rw.prior)

                if (rw.parent is not None and i == 0):
                    rw.parent.mid_do(rw)

                await rw.parent.finish_item(bus_req)
                await bus_req.end_event.wait_on()

                if adapter.provides_responses:
                    bus_rsp = None  # uvm_sequence_item
                    op = 0  # uvm_access_e
                    # TODO: need to test for right trans type, if not put back in q
                    await rw.parent.get_base_response(bus_rsp)
                    rw_access = adapter.bus2reg(bus_rsp,rw_access)
                else:
                    adapter.bus2reg(bus_req,rw_access)

                data = rw_access.data & ((1 << bus_width*8)-1)  # mask the upper bits
                rw.status = rw_access.status

                # TODO
                #if (rw.status == UVM_IS_OK && (^data) === 1'bx):

                uvm_info(self.get_type_name(),
                   sv.sformatf("Read 0x%h at 0x%h via map %s: %s...", data,
                       addrs[i], self.get_full_name(), str(rw.status)), UVM_VERB_MEM_MAP)

                if (rw.status == UVM_NOT_OK):
                    break

                rw.value[val_idx] |= data << curr_byte_*8

                if (rw.parent is not None and i == len(addrs)-1):
                    rw.parent.post_do(rw)

            #foreach (addrs[i])
            for i in range(len(addrs)):
                addrs[i] = addrs[i] + map_info.mem_range.stride

            if (rw.element_kind == UVM_FIELD):
                rw.value[val_idx] = (rw.value[val_idx] >> (n_access_extra)) & ((1<<size)-1)
        #endtask: do_bus_read


    async def do_write(self, rw):
        """
        Perform a write operation.

        Args:
            rw (UVMRegItem): Register item for the write op.
        """
        tmp_parent_seq = None  # uvm_sequence_base
        system_map = self.get_root_map()
        adapter = system_map.get_adapter()
        sequencer = system_map.get_sequencer()

        if (adapter is not None and adapter.parent_sequence is not None):
            o = None
            seq = None
            o = adapter.parent_sequence.clone()
            #assert($cast(seq,o))
            seq = o
            seq.set_parent_sequence(rw.parent)
            rw.parent = seq
            tmp_parent_seq = seq

        if rw.parent is None:
            #rw.parent = new("default_parent_seq")
            rw.parent = UVMSequenceBase("default_parent_seq")
            tmp_parent_seq = rw.parent

        if adapter is None:
            uvm_debug(self, 'do_write', 'Start seq because adapter is None')
            rw.set_sequencer(sequencer)
            await rw.parent.start_item(rw,rw.prior)
            await rw.parent.finish_item(rw)
            await rw.end_event.wait_on()
        else:
            uvm_debug(self, 'do_write', 'do_bus_write because adapter exists')
            await self.do_bus_write(rw, sequencer, adapter)

        # TODO
        #  if (tmp_parent_seq is not None)
        #    sequencer.m_sequence_exiting(tmp_parent_seq)
        #
        #endtask


    async def do_read(self, rw):
        """
        Perform a read operation.

        Args:
            rw (UVMRegItem): Register item for the read op.
        """
        tmp_parent_seq = None  # uvm_sequence_base
        system_map = self.get_root_map()
        adapter = system_map.get_adapter()
        sequencer = system_map.get_sequencer()

        if (adapter is not None and adapter.parent_sequence is not None):
            o = None  # uvm_object
            seq = None  # uvm_sequence_base
            o = adapter.parent_sequence.clone()
            # assert($cast(seq,o)) TODO check this
            seq = o
            seq.set_parent_sequence(rw.parent)
            rw.parent = seq
            tmp_parent_seq = seq

        if rw.parent is None:
            rw.parent = UVMSequenceBase("default_parent_seq")
            tmp_parent_seq = rw.parent

        if adapter is None:
            rw.set_sequencer(sequencer)
            await rw.parent.start_item(rw, rw.prior)
            await rw.parent.finish_item(rw)
            await rw.end_event.wait_on()
        else:
            await self.do_bus_read(rw, sequencer, adapter)

        if tmp_parent_seq is not None:
            sequencer.m_sequence_exiting(tmp_parent_seq)



    def Xget_bus_infoX(self, rw, map_info, size, lsb, addr_skip):
        """
        Returns bus info based on given register item.

        Args:
            rw (UVMRegItem):
            map_info (UVMRegMapInfo):
            size (int):
            lsb (int):
            addr_skip (int):
        Returns:
        Raises:
        """
        if rw.element_kind == UVM_MEM:
            mem = None  # uvm_mem
            if rw.element is None:  # || !$cast(mem,rw.element)):
                uvm_fatal("REG/CAST", "uvm_reg_item 'element_kind' is UVM_MEM, " +
                    "but 'element' does not point to a memory: " + rw.get_name())
            mem = rw.element
            map_info = self.get_mem_map_info(mem)
            size = mem.get_n_bits()
        elif rw.element_kind == UVM_REG:
            rg = None  # uvm_reg
            if(rw.element is None):  # || !$cast(rg,rw.element))
                uvm_fatal("REG/CAST", "uvm_reg_item 'element_kind' is UVM_REG, " +
                    "but 'element' does not point to a register: " + rw.get_name())
            rg = rw.element
            map_info = self.get_reg_map_info(rg)
            size = rg.get_n_bits()
        elif rw.element_kind == UVM_FIELD:
            field = None  # uvm_reg_field
            if rw.element is None:  # || !$cast(field,rw.element))
                uvm_fatal("REG/CAST", "uvm_reg_item 'element_kind' is UVM_FIELD, " +
                    "but 'element' does not point to a field: " + rw.get_name())
            field = rw.element
            map_info = self.get_reg_map_info(field.get_parent())
            size = field.get_n_bits()
            lsb = field.get_lsb_pos()
            addr_skip = (int(lsb/(self.get_n_bytes())*8))
        else:
            raise Exception("rw.element_kind value illegal: " +
                    str(rw.element_kind))
        return [map_info, size, lsb, addr_skip]


    #   extern virtual function string      convert2string()

    #   extern virtual function uvm_object  clone()
    #   extern virtual function void        do_print (uvm_printer printer)
    #   extern virtual function void        do_copy   (uvm_object rhs)
    #   //extern virtual function bit       do_compare (uvm_object rhs, uvm_comparer comparer)
    #   //extern virtual function void      do_pack (uvm_packer packer)
    #   //extern virtual function void      do_unpack (uvm_packer packer)
    #
    #
    #    // Function: set_transaction_order_policy
    #    // set the transaction order policy
    #    function void set_transaction_order_policy(uvm_reg_transaction_order_policy pol)
    #        policy = pol
    #    endfunction
    #
    #    // Function: get_transaction_order_policy
    #    // set the transaction order policy
    #    function uvm_reg_transaction_order_policy get_transaction_order_policy()
    #        return policy
    #    endfunction
    #
    #endclass: uvm_reg_map
uvm_object_utils(UVMRegMap)

#---------------
# Initialization
#---------------
#
#
# m_set_mem_offset
#
#function void uvm_reg_map::m_set_mem_offset(uvm_mem mem,
#                                            uvm_reg_addr_t offset,
#                                            bit unmapped)
#
#   if (!self.m_mems_info.exists(mem)):
#      `uvm_error("RegModel",
#         {"Cannot modify offset of memory '",mem.get_full_name(),
#         "' in address map '",get_full_name(),
#         "' : memory not mapped in that address map"})
#      return
#   end
#
#   begin
#      UVMRegMapInfo info    = self.m_mems_info[mem]
#      uvm_reg_block    blk     = get_parent()
#      uvm_reg_map      top_map = get_root_map()
#      uvm_reg_addr_t   addrs[]
#
#      // if block is not locked, Xinit_address_mapX will resolve map when block is locked
#      if (blk.is_locked()):
#
#         // remove any existing cached addresses
#         if (!info.unmapped):
#           foreach (top_map.m_mems_by_offset[range]):
#              if (top_map.m_mems_by_offset[range] == mem)
#                 top_map.m_mems_by_offset.delete(range)
#           end
#         end
#
#         // if we are remapping...
#         if (!unmapped):
#            uvm_reg_addr_t addrs[],addrs_max[]
#            uvm_reg_addr_t min, max, min2, max2
#            int unsigned stride
#
#            void'(get_physical_addresses(offset,0,mem.get_n_bytes(),addrs))
#            min = (addrs[0] < addrs[addrs.size()-1]) ? addrs[0] : addrs[addrs.size()-1]
#            min2 = addrs[0]
#
#            void'(get_physical_addresses(offset,(mem.get_size()-1),
#                                         mem.get_n_bytes(),addrs_max))
#            max = (addrs_max[0] > addrs_max[addrs_max.size()-1]) ?
#               addrs_max[0] : addrs_max[addrs_max.size()-1]
#            max2 = addrs_max[0]
#            // address interval between consecutive mem locations
#            stride = (max2 - max)/(mem.get_size()-1)
#
#            // make sure new offset does not conflict with others
#            foreach (top_map.m_regs_by_offset[reg_addr]):
#               if (reg_addr >= min && reg_addr <= max):
#                  string a,b
#                  a = $sformatf("[%0h:%0h]",min,max)
#                  b = $sformatf("%0h",reg_addr)
#                  `uvm_warning("RegModel", {"In map '",get_full_name(),"' memory '",
#                      mem.get_full_name(), "' with range ",a,
#                      " overlaps with address of existing register '",
#                      top_map.m_regs_by_offset[reg_addr].get_full_name(),"': 'h",b})
#               end
#            end
#
#            foreach (top_map.m_mems_by_offset[range]):
#               if (min <= range.max && max >= range.max ||
#                   min <= range.min && max >= range.min ||
#                   min >= range.min && max <= range.max):
#                 string a,b
#                 a = $sformatf("[%0h:%0h]",min,max)
#                 b = $sformatf("[%0h:%0h]",range.min,range.max)
#                 `uvm_warning("RegModel", {"In map '",get_full_name(),"' memory '",
#                     mem.get_full_name(), "' with range ",a,
#                     " overlaps existing memory with range '",
#                     top_map.m_mems_by_offset[range].get_full_name(),"': ",b})
#                 end
#            end
#
#            begin
#              UVMRegMapAddrRange range = '{ min, max, stride }
#              top_map.m_mems_by_offset[range] = mem
#              info.addr  = addrs
#              info.mem_range = range
#            end
#
#         end
#      end
#
#      if (unmapped):
#        info.offset   = -1
#        info.unmapped = 1
#      end
#      else begin
#        info.offset   = offset
#        info.unmapped = 0
#      end
#
#   end
#endfunction
#
#
#
#------------
# get methods
#------------
#
#
#
# get_fields
#
#function void uvm_reg_map::get_fields(ref uvm_reg_field fields[$], input uvm_hier_e hier=UVM_HIER)
#
#   foreach (self.m_regs_info[rg_]):
#     uvm_reg rg = rg_
#     rg.get_fields(fields)
#   end
#
#   if (hier == UVM_HIER)
#     foreach (this.self.m_submaps[submap_]):
#       uvm_reg_map submap=submap_
#       submap.get_fields(fields)
#     end
#
#endfunction
#
#
# get_memories
#
#function void uvm_reg_map::get_memories(ref uvm_mem mems[$], input uvm_hier_e hier=UVM_HIER)
#
#   foreach (self.m_mems_info[mem])
#     mems.push_back(mem)
#
#   if (hier == UVM_HIER)
#     foreach (self.m_submaps[submap_]):
#       uvm_reg_map submap=submap_
#       submap.get_memories(mems)
#     end
#
#endfunction
#
#
# get_virtual_registers
#
#function void uvm_reg_map::get_virtual_registers(ref uvm_vreg regs[$], input uvm_hier_e hier=UVM_HIER)
#
#  uvm_mem mems[$]
#  get_memories(mems,hier)
#
#  foreach (mems[i])
#    mems[i].get_virtual_registers(regs)
#
#endfunction
#
#
# get_virtual_fields
#
#function void uvm_reg_map::get_virtual_fields(ref uvm_vreg_field fields[$], input uvm_hier_e hier=UVM_HIER)
#
#   uvm_vreg regs[$]
#   get_virtual_registers(regs,hier)
#
#   foreach (regs[i])
#       regs[i].get_fields(fields)
#
#endfunction
#
#
#----------
# Size and Overlap Detection
#---------
#
# set_base_addr
#
#function void uvm_reg_map::set_base_addr(uvm_reg_addr_t offset)
#   if (self.m_parent_map is not None):
#      self.m_parent_map.set_submap_offset(this, offset)
#   end
#   else begin
#      self.m_base_addr = offset
#      if (self.m_parent.is_locked()):
#         uvm_reg_map top_map = get_root_map()
#         top_map.Xinit_address_mapX()
#      end
#   end
#endfunction
#
#
# get_size
#
#function int unsigned uvm_reg_map::get_size()
#
#  int unsigned max_addr
#  int unsigned addr
#
#  // get max offset from registers
#  foreach (self.m_regs_info[rg_]):
#    uvm_reg rg = rg_
#    addr = self.m_regs_info[rg].offset + ((rg.get_n_bytes()-1)/self.m_n_bytes)
#    if (addr > max_addr) max_addr = addr
#  end
#
#  // get max offset from memories
#  foreach (self.m_mems_info[mem_]):
#    uvm_mem mem = mem_
#    addr = self.m_mems_info[mem].offset + (mem.get_size() * (((mem.get_n_bytes()-1)/self.m_n_bytes)+1)) -1
#    if (addr > max_addr) max_addr = addr
#  end
#
#  // get max offset from submaps
#  foreach (self.m_submaps[submap_]):
#    uvm_reg_map submap=submap_
#    addr = self.m_submaps[submap] + submap.get_size()
#    if (addr > max_addr) max_addr = addr
#  end
#
#  return max_addr + 1
#
#endfunction
#
#
#
#function void uvm_reg_map::Xverify_map_configX()
#   // Make sure there is a generic payload sequence for each map
#   // in the model and vice-versa if this is a root sequencer
#   bit error
#   uvm_reg_map root_map = get_root_map()
#
#   if (root_map.get_adapter() is None):
#      `uvm_error("RegModel", {"Map '",root_map.get_full_name(),
#                 "' does not have an adapter registered"})
#      error++
#   end
#   if (root_map.get_sequencer() is None):
#      `uvm_error("RegModel", {"Map '",root_map.get_full_name(),
#                 "' does not have a sequencer registered"})
#      error++
#   end
#   if (error):
#      `uvm_fatal("RegModel", {"Must register an adapter and sequencer ",
#                 "for each top-level map in RegModel model"})
#      return
#   end
#
#endfunction
#
#
#--------------
# Get-By-Offset
#--------------
#
# get_mem_by_offset
#
#function uvm_mem uvm_reg_map::get_mem_by_offset(uvm_reg_addr_t offset)
#   if (!self.m_parent.is_locked()):
#      `uvm_error("RegModel", $sformatf("Cannot memory register by offset: Block %s is not locked.", self.m_parent.get_full_name()))
#      return None
#   end
#
#   foreach (self.m_mems_by_offset[range]):
#      if (range.min <= offset && offset <= range.max):
#         return self.m_mems_by_offset[range]
#      end
#   end
#
#   return None
#endfunction
#
#-------------
# Standard Ops
#-------------
#
# do_print
#
#function void uvm_reg_map::do_print (uvm_printer printer)
#   uvm_reg  regs[$]
#   uvm_vreg vregs[$]
#   uvm_mem  mems[$]
#   uvm_endianness_e endian
#   uvm_reg_map maps[$]
#   string prefix
#   uvm_sequencer_base sqr=get_sequencer()
#
#   super.do_print(printer)
#  printer.print_generic(get_name(), self.get_type_name(), -1, convert2string())
#
#   endian = get_endian(UVM_NO_HIER)
#   $sformat(convert2string, "%s -- %0d bytes (%s)", convert2string,
#            get_n_bytes(UVM_NO_HIER), endian.name())
#
#   printer.print_generic("endian","",-2,endian.name())
#   if(sqr is not None)
#    printer.print_generic("effective sequencer",sqr.get_type_name(),-2,sqr.get_full_name())
#
#   get_registers(regs,UVM_NO_HIER)
#   foreach (regs[j])
#        printer.print_generic(regs[j].get_name(), regs[j].get_type_name(),-2,$sformatf("@%0d +'h%0x",regs[j].get_inst_id(),regs[j].get_address(this)))
#
#
#   get_memories(mems)
#   foreach (mems[j])
#        printer.print_generic(mems[j].get_name(), mems[j].get_type_name(),-2,$sformatf("@%0d +'h%0x",mems[j].get_inst_id(),mems[j].get_address(0,this)))
#
#   get_virtual_registers(vregs)
#   foreach (vregs[j])
#        printer.print_generic(vregs[j].get_name(), vregs[j].get_type_name(),-2,$sformatf("@%0d +'h%0x",vregs[j].get_inst_id(),vregs[j].get_address(0,this)))
#
#   get_submaps(maps)
#   foreach (maps[j])
#        printer.print_object(maps[j].get_name(),maps[j])
#endfunction
#
# convert2string
#
#function string uvm_reg_map::convert2string()
#   uvm_reg  regs[$]
#   uvm_vreg vregs[$]
#   uvm_mem  mems[$]
#   uvm_endianness_e endian
#   string prefix
#
#   $sformat(convert2string, "%sMap %s", prefix, get_full_name())
#   endian = get_endian(UVM_NO_HIER)
#   $sformat(convert2string, "%s -- %0d bytes (%s)", convert2string,
#            get_n_bytes(UVM_NO_HIER), endian.name())
#   get_registers(regs)
#   foreach (regs[j]):
#      $sformat(convert2string, "%s\n%s", convert2string,
#               regs[j].convert2string());//{prefix, "   "}, this))
#   end
#   get_memories(mems)
#   foreach (mems[j]):
#      $sformat(convert2string, "%s\n%s", convert2string,
#               mems[j].convert2string());//{prefix, "   "}, this))
#   end
#   get_virtual_registers(vregs)
#   foreach (vregs[j]):
#      $sformat(convert2string, "%s\n%s", convert2string,
#               vregs[j].convert2string());//{prefix, "   "}, this))
#   end
#endfunction
#
#
# clone
#
#function uvm_object uvm_reg_map::clone()
#  //uvm_rap_map me
#  //me = new this
#  //return me
#  return None
#endfunction
#
#
# do_copy
#
#function void uvm_reg_map::do_copy (uvm_object rhs)
#  //uvm_reg_map rhs_
#  //assert($cast(rhs_,rhs))
#
#  //rhs_.regs = regs
#  //rhs_.mems = mems
#  //rhs_.vregs = vregs
#  //rhs_.blks = blks
#  //... and so on
#endfunction



#//
#//-----------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2013 NVIDIA Corporation
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//-----------------------------------------------------------------------------
"""
File: UVM Links

The <UVMLinkBase> class, and its extensions, are provided as a mechanism
to allow for compile-time safety when trying to establish links between
records within a <uvm_tr_database>.
"""



from .uvm_object import UVMObject
from ..macros import uvm_object_utils


class UVMLinkBase(UVMObject):
    """

    CLASS: UVMLinkBase

    The ~UVMLinkBase~ class presents a simple API for defining a link between
    any two objects.

    Using extensions of self class, a <uvm_tr_database> can determine the
    type of links being passed, without relying on "magic" string names.

    For example::

       def do_establish_link(self, link):
         pc_link: UVMParentChildLink = None
         ce_link: UVMCauseEffectLink = None

         if (sv.cast(pc_link, link)):
            # Record the parent-child relationship
         elif (sv.cast(ce_link, link)):
            # Record the cause-effect relationship
         else:
            # Unsupported relationship!

    """


    def __init__(self, name="unnamed-UVMLinkBase"):
        """
        Function: new
        Constructor

        Args:
            name (str): Name of the link.
        """
        super().__init__(name)

    def set_lhs(self, lhs: UVMObject):
        """
        Group:  Accessors

        Function: set_lhs
        Sets the left-hand-side of the link

        Triggers the `do_set_lhs` callback.

        Args:
            lhs:
        """
        self.do_set_lhs(lhs)

    def get_lhs(self) -> UVMObject:
        """
        Function: get_lhs
        Gets the left-hand-side of the link

        Triggers the `do_get_lhs` callback

        Returns:
        """
        return self.do_get_lhs()

    def set_rhs(self, rhs: UVMObject) -> None:
        """
           Function: set_rhs
           Sets the right-hand-side of the link

           Triggers the `do_set_rhs` callback.
        Args:
            rhs:
        """
        self.do_set_rhs(rhs)


    def get_rhs(self) -> UVMObject:
        """
           Function: get_rhs
           Gets the right-hand-side of the link

           Triggers the `do_get_rhs` callback
        Returns:
        """
        return self.do_get_rhs()


    def set(self, lhs: UVMObject, rhs: UVMObject) -> None:
        """
           Function: set
           Convenience method for setting both sides in one call.

           Triggers both the `do_set_rhs` and `do_set_lhs` callbacks.
        Args:
            lhs:
            rhs:
        """
        self.do_set_lhs(lhs)
        self.do_set_rhs(rhs)


    #   // Group: Implementation Callbacks

    #   // Function: do_set_lhs
    #   // Callback for setting the left-hand-side
    def do_set_lhs(self, lhs: UVMObject) -> None:
        raise NotImplementedError('Pure virtual function')

    #   // Function: do_get_lhs
    #   // Callback for retrieving the left-hand-side
    def do_get_lhs(self):
        raise NotImplementedError('Pure virtual function')

    #   // Function: do_set_rhs
    #   // Callback for setting the right-hand-side
    def do_set_rhs(self, rhs: UVMObject) -> None:
        raise NotImplementedError('Pure virtual function')

    #   // Function: do_get_rhs
    #   // Callback for retrieving the right-hand-side
    def do_get_rhs(self):
        raise NotImplementedError('Pure virtual function')




class UVMParentChildLink(UVMLinkBase):
    """
    CLASS: UVMParentChildLink

    The ~UVMParentChildLink~ is used to represent a Parent/Child relationship
    between two objects.
    """


    def __init__(self, name="unnamed-UVMParentChildLink"):
        """
           Function: new
           Constructor

           Parameters:
           name - Instance name
        Args:
            name:
        """
        super().__init__(name)
        self.m_lhs = None
        self.m_rhs = None


    @classmethod
    def get_link(cls, lhs: UVMObject, rhs: UVMObject, name="pc_link") -> 'UVMParentChildLink':
        """
           Function: get_link
           Constructs a pre-filled link

           This allows for simple one-line link creations.
        .. code-block:: python

           | my_db.establish_link(UVMParentChildLink::get_link(record1, record2))

           Parameters:
           lhs - Left hand side reference
           rhs - Right hand side reference
           name - Optional name for the link object

        Args:
            cls:
            lhs:
            rhs:
            name:
        Returns:
        """
        pc_link = UVMParentChildLink(name)
        pc_link.set(lhs, rhs)
        return pc_link


    def do_set_lhs(self, lhs):
        """
           Group: Implementation Callbacks

           Function: do_set_lhs
           Sets the left-hand-side (Parent)

        Args:
            lhs:
        """
        self.m_lhs = lhs

    def do_get_lhs(self):
        """
           Function: do_get_lhs
           Retrieves the left-hand-side (Parent)

        Returns:
        """
        return self.m_lhs

    def do_set_rhs(self, rhs):
        """
           Function: do_set_rhs
           Sets the right-hand-side (Child)

        Args:
            rhs:
        """
        self.m_rhs = rhs


    def do_get_rhs(self):
        """
           Function: do_get_rhs
           Retrieves the right-hand-side (Child)

        Returns:
        """
        return self.m_rhs


class UVMCauseEffectLink(UVMLinkBase):
    """

    CLASS: UVMCauseEffectLink

    The ~UVMCauseEffectLink~ is used to represent a Cause/Effect relationship
    between two objects.
    """




    def __init__(self, name="unnamed-UVMCauseEffectLink"):
        """
           Function: new
           Constructor

           Parameters:
           name - Instance name
        Args:
            name:
        """
        super().__init__(name)
        self.m_lhs: UVMObject = None
        self.m_rhs: UVMObject = None


    @classmethod
    def get_link(cls, lhs, rhs, name="ce_link") -> 'UVMCauseEffectLink':
        """
           Function: get_link
           Constructs a pre-filled link

           This allows for simple one-line link creations.
        .. code-block:: python

           | my_db.establish_link(UVMCauseEffectLink::get_link(record1, record2))

           Parameters:
           lhs - Left hand side reference
           rhs - Right hand side reference
           name - Optional name for the link object

        Args:
            cls:
            lhs:
            rhs:
            name:
        Returns:
        """
        ce_link = UVMCauseEffectLink(name)
        ce_link.set(lhs, rhs)
        return ce_link


    def do_set_lhs(self, lhs):
        """
           Group: Implementation Callbacks

           Function: do_set_lhs
           Sets the left-hand-side (Cause)

        Args:
            lhs:
        """
        self.m_lhs = lhs


    def do_get_lhs(self):
        """
           Function: do_get_lhs
           Retrieves the left-hand-side (Cause)

        Returns:
        """
        return self.m_lhs


    def do_set_rhs(self, rhs):
        """
           Function: do_set_rhs
           Sets the right-hand-side (Effect)

        Args:
            rhs:
        """
        self.m_rhs = rhs


    def do_get_rhs(self):
        """
           Function: do_get_rhs
           Retrieves the right-hand-side (Effect)

        Returns:
        """
        return self.m_rhs


uvm_object_utils(UVMParentChildLink)


class UVMRelatedLink(UVMLinkBase):
    """
    CLASS: UVMRelatedLink

    The ~UVMRelatedLink~ is used to represent a generic "is related" link
    between two objects.
    """


    def __init__(self, name="unnamed-UVMRelatedLink"):
        """
           Function: new
           Constructor

           Parameters:
           name - Instance name
        Args:
            name:
        """
        super().__init__(name)
        self.m_lhs = None  # type: UVMObject
        self.m_rhs = None  # type: UVMObject


    @classmethod
    def get_link(cls, lhs, rhs, name="ce_link") -> 'UVMRelatedLink':
        """
           Function: get_link
           Constructs a pre-filled link

           This allows for simple one-line link creations.
        .. code-block:: python

           | my_db.establish_link(UVMRelatedLink::get_link(record1, record2))

        Args:
            lhs: Left hand side reference
            rhs: Right hand side reference
            name (str): Optional name for the link object

        Returns:
            UVMRelatedLink: Created related link.
        """
        ce_link = UVMRelatedLink(name)
        ce_link.set(lhs, rhs)
        return ce_link


    #   // Group: Implementation Callbacks
    #
    #   // Function: do_set_lhs
    #   // Sets the left-hand-side
    #   //
    #   virtual def void do_set_lhs(self,UVMObject lhs):
    #      m_lhs = lhs
    #   endfunction : do_set_lhs

    #   // Function: do_get_lhs
    #   // Retrieves the left-hand-side
    #   //
    #   virtual def UVMObject do_get_lhs(self):
    #      return m_lhs
    #   endfunction : do_get_lhs

    #   // Function: do_set_rhs
    #   // Sets the right-hand-side
    #   //
    #   virtual def void do_set_rhs(self,UVMObject rhs):
    #      m_rhs = rhs
    #   endfunction : do_set_rhs

    #   // Function: do_get_rhs
    #   // Retrieves the right-hand-side
    #   //
    #   virtual def UVMObject do_get_rhs(self):
    #      return m_rhs
    #   endfunction : do_get_rhs


uvm_object_utils(UVMRelatedLink)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

from uvm import (UVMEnv, uvm_component_utils, UVMRegPredictor)
from common.reg_agent import (reg_agent, reg2rw_adapter)
from regmodel import block_B


class dut():
    #   static bit [3:0] Ra_F1 = 0
    #   static bit [3:0] Ra_F2 = 0
    #   static bit [3:0] Rb_F1 = 0
    #   static bit [3:0] Rb_F2 = 0
    #   static bit [7:0] M[1024]

    @classmethod
    def rw(cls, rw):
        pass
        #      casez (rw.addr)
        #        0x0000: // Ra
        #          if (rw.read) rw.data = {16'h00, Ra_F2, Ra_F1}
        #          else begin
        #             if (rw.byte_en[0]) Ra_F1 = rw.data[ 7:0]
        #             if (rw.byte_en[0]) Ra_F2 = rw.data[15:8]
        #          end
        #        0x0100: // Rb
        #          if (rw.read) rw.data = {16'h00, Rb_F2, Rb_F1}
        #          else begin
        #             if (rw.byte_en[0]) Rb_F1 = rw.data[ 7:0]
        #             if (rw.byte_en[0]) Rb_F2 = rw.data[15:8]
        #          end
        #        0b001000??_????????: // M
        #          if (rw.read) rw.data = {24'h00, M[rw.addr[9:0]]}
        #          else begin
        #             if (rw.byte_en[0]) M[rw.addr[9:0]] = rw.data[ 7:0]
        #          end
        #      endcase
        #      yield Timer(1, )
        #      // $write("%s\n", rw.convert2string());
        #   endtask
        #endclass


class tb_env(UVMEnv):

    #   block_B   regmodel
    #   reg_agent#(dut) bus
    #   uvm_reg_predictor#(reg_rw) predict

    def __init__(self, name="tb_env", parent=None):
        super().__init__(name, parent)
        self.regmodel = None
        self.bus = None
        self.predict = None

    def build_phase(self, phase):
        self.regmodel = block_B.type_id.create("regmodel")
        self.regmodel.build()
        self.regmodel.lock_model()

        self.bus = reg_agent.type_id.create("bus", self)
        self.bus.set_nbits(16)
        self.predict = UVMRegPredictor.type_id.create("predict", self)


    def connect_phase(self, phase):
        self.reg2rw  = reg2rw_adapter("reg2rw")
        self.regmodel.default_map.set_sequencer(self.bus.sqr, self.reg2rw)

        self.predict.map = self.regmodel.default_map
        self.predict.adapter = self.reg2rw
        self.bus.mon.ap.connect(self.predict.bus_in)
        self.regmodel.default_map.set_auto_predict(0)

    #endclass


uvm_component_utils(tb_env)



import unittest

from uvm.base.uvm_coreservice import UVMCoreService


class TestUVMCoreService(unittest.TestCase):

    def test_name(self):
        cs = UVMCoreService()
        root = cs.get_root()
        self.assertEqual(root.get_full_name(), "")


if __name__ == '__main__':
    unittest.main()



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------
"""
Title: Comparators

The following classes define comparators for objects and built-in types.
"""

from ..base.uvm_component import UVMComponent
from ..base.uvm_object_globals import UVM_MEDIUM
from ..base.sv import sv
from ..macros import uvm_component_utils, uvm_error, uvm_info
from ..tlm1 import UVMAnalysisExport, UVMAnalysisPort, UVMTLMAnalysisFIFO
from .uvm_pair import UVMBuiltInPair, UVMClassPair
from .uvm_policies import (UVMBuiltInConverter, UVMBuiltInComp,
    UVMClassConverter, UVMClassComp)


class UVMInOrderComparator(UVMComponent):
    """
    CLASS: UVMInOrderComparator #(T,comp_type,convert,pair_type)

    Compares two streams of data objects of the type parameter, T.
    These transactions may either be classes or built-in types. To be
    successfully compared, the two streams of data must be in the same order.
    Apart from that, there are no assumptions made about the relative timing of
    the two streams of data.

    Type parameters

      T       - Specifies the type of transactions to be compared.

      comp_type - A policy class to compare the two
                  transaction streams. It must provide the static method
                  "function bit comp(T a, T b)" which returns ~TRUE~
                  if ~a~ and ~b~ are the same.

      convert - A policy class to convert the transactions being compared
                to a string. It must provide the static method
                "function string convert2string(T a)".

     pair_type - A policy class to allow pairs of transactions to be handled as
                 a single <uvm_object> type.

    Built in types (such as ints, bits, logic, and structs) can be compared using
    the default values for comp_type, convert, and pair_type. For convenience,
    you can use the subtype, <UVMInOrderBuiltInComparator #(T)>
    for built-in types.

    When T is a `UVMObject`, you can use the convenience subtype
    `UVMInOrderClassComparator`.

    Comparisons are commutative, meaning it does not matter which data stream is
    connected to which export, before_export or after_export.

    Comparisons are done in order and as soon as a transaction is received from
    both streams. Internal fifos are used to buffer incoming transactions on one
    stream until a transaction to compare arrives on the other stream.
    """


    #  #( type T = int ,
    #     type comp_type = uvm_built_in_comp #( T ) ,
    #     type convert = uvm_built_in_converter #( T ) ,
    #     type pair_type = uvm_built_in_pair #( T ) )
    #    extends uvm_component
    #

    type_name = "UVMInOrderComparator"

    #  // Port: before_export
    #  //
    #  // The export to which one stream of data is written. The port must be
    #  // connected to an analysis port that will provide such data.
    #
    #  uvm_analysis_export #(T) before_export


    #  // Port: after_export
    #  //
    #  // The export to which the other stream of data is written. The port must be
    #  // connected to an analysis port that will provide such data.
    #
    #  uvm_analysis_export #(T) after_export


    #  // Port: pair_ap
    #  //
    #  // The comparator sends out pairs of transactions across this analysis port.
    #  // Both matched and unmatched pairs are published via a pair_type objects.
    #  // Any connected analysis export(s) will receive these transaction pairs.
    #
    #  uvm_analysis_port   #(pair_type) pair_ap


    def __init__(self, name, parent):
        super().__init__(name, parent)

        self.PairType = UVMBuiltInPair
        self.Convert = UVMBuiltInConverter
        self.CompType = UVMBuiltInComp

        self.before_export = UVMAnalysisExport("before_export", self)
        self.after_export  = UVMAnalysisExport("after_export", self)
        self.pair_ap       = UVMAnalysisPort("pair_ap", self)

        self.m_before_fifo = UVMTLMAnalysisFIFO("before", self)
        self.m_after_fifo  = UVMTLMAnalysisFIFO("after", self)
        self.m_matches = 0
        self.m_mismatches = 0


    def get_type_name(self):
        return UVMInOrderComparator.type_name


    def connect_phase(self, phase):
        self.before_export.connect(self.m_before_fifo.analysis_export)
        self.after_export.connect(self.m_after_fifo.analysis_export)



    #  // Task- run_phase
    #  //
    #  // Internal method.
    #  //
    #  // Takes pairs of before and after transactions and compares them.
    #  // Status information is updated according to the results of the comparison.
    #  // Each pair is published to the pair_ap analysis port.
    async def run_phase(self, phase):

        #    pair_type pair
        #    T b
        #    T a

        s = ""
        await super().run_phase(phase)  # Is this needed?
        while True:

            arr_b = []
            arr_a = []

            await self.m_before_fifo.get(arr_b)
            await self.m_after_fifo.get(arr_a)

            b = arr_b[0]
            a = arr_a[0]

            # pytype: disable=wrong-arg-types
            if self.CompType.comp(b, a) is False:
                s = sv.sformatf("%s differs from %s", self.Convert.convert2string(a),
                    self.Convert.convert2string(b))
                uvm_error("Comparator Mismatch", s)
                self.m_mismatches += 1

            else:
                s = self.Convert.convert2string(b)
                uvm_info("Comparator Match", s, UVM_MEDIUM)
                self.m_matches += 1

            # we make the assumption here that a transaction "sent for
            # analysis" is safe from being edited by another process.
            # Hence, it is safe not to clone a and b.

            pair = self.PairType("after/before")
            # pytype: enable=wrong-arg-types
            pair.first = a
            pair.second = b
            self.pair_ap.write(pair)


    #  // Function: flush
    #  //
    #  // This method sets m_matches and m_mismatches back to zero. The
    #  // `UVMTLMFIFO.flush` takes care of flushing the FIFOs.
    def flush(self):
        self.m_matches = 0
        self.m_mismatches = 0


uvm_component_utils(UVMInOrderComparator)


class UVMInOrderBuiltInComparator(UVMInOrderComparator):
    """
    CLASS: UVMInOrderBuiltInComparator

    This class uses the uvm_built_in_* comparison, converter, and pair classes.
    Use this class for built-in types (int, bit, string, etc.)
    """


    type_name = "UVMInOrderBuiltInComparator"

    def __init__(self, name, parent):
        super().__init__(name, parent)


    def get_type_name(self):
        return UVMInOrderBuiltInComparator.type_name


uvm_component_utils(UVMInOrderBuiltInComparator)


class UVMInOrderClassComparator(UVMInOrderComparator):
    """
    CLASS: UVMInOrderClassComparator #(T)

    This class uses the uvm_class_* comparison, converter, and pair classes.
    Use this class for comparing user-defined objects of type T, which must
    provide compare() and convert2string() method.
    """

    type_name = "UVMInOrderClassComparator"

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.PairType = UVMClassPair
        self.Convert = UVMClassConverter
        self.CompType = UVMClassComp


    def get_type_name(self):
        return UVMInOrderClassComparator.type_name


uvm_component_utils(UVMInOrderClassComparator)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2009 Synopsys, Inc.
#//    Copyright 2010-2011 Mentor Graphics Corporation
#//    Copyright 2010-2011 Cadence Design Systems, Inc.
#//    Copyright 2019 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import Timer
from ..seq.uvm_sequence import UVMSequence
from ..macros import uvm_object_utils, uvm_fatal, uvm_warning, uvm_info
from ..base import UVM_LOW, UVM_HIGH
from ..base.uvm_globals import uvm_zero_delay
from .uvm_reg_model import UVM_WRITE

#//------------------------------------------------------------------------------
#// TITLE: Register Sequence Classes
#//------------------------------------------------------------------------------
#//
#// This section defines the base classes used for register stimulus generation.
#//------------------------------------------------------------------------------

LOCAL = 0
UPSTREAM = 1

#//------------------------------------------------------------------------------
#//
#// CLASS: UVMRegSequence
#//
#// This class provides base functionality for both user-defined RegModel test
#// sequences and "register translation sequences".
#//
#// - When used as a base for user-defined RegModel test sequences, this class
#//   provides convenience methods for reading and writing registers and
#//   memories. Users implement the body() method to interact directly with
#//   the RegModel model (held in the <model> property) or indirectly via the
#//   delegation methods in this class.
#//
#// - When used as a translation sequence, objects of this class are
#//   executed directly on a bus sequencer which are used in support of a layered sequencer
#//   use model, a pre-defined convert-and-execute algorithm is provided.
#//
#// Register operations do not require extending this class if none of the above
#// services are needed. Register test sequences can be extend from the base
#// <uvm_sequence #(REQ,RSP)> base class or even from outside a sequence.
#//
#// Note- The convenience API not yet implemented.
#//------------------------------------------------------------------------------
#class UVMRegSequence #(type BASE=uvm_sequence #(uvm_reg_item)) extends BASE;


class UVMRegSequence(UVMSequence):
    #
    #
    #  // Parameter: BASE
    #  //
    #  // Specifies the sequence type to extend from.
    #  //
    #  // When used as a translation sequence running on a bus sequencer, ~BASE~ must
    #  // be compatible with the sequence type expected by the bus sequencer.
    #  //
    #  // When used as a test sequence running on a particular sequencer, ~BASE~
    #  // must be compatible with the sequence type expected by that sequencer.
    #  //
    #  // When used as a virtual test sequence without a sequencer, ~BASE~ does
    #  // not need to be specified, i.e. the default specialization is adequate.
    #  //
    #  // To maximize opportunities for reuse, user-defined RegModel sequences should
    #  // "promote" the BASE parameter.
    #  //
    #  // | class my_reg_sequence #(type BASE=uvm_sequence #(uvm_reg_item))
    #  // |                               extends UVMRegSequence #(BASE);
    #  //
    #  // This way, the RegModel sequence can be extended from
    #  // user-defined base sequences.

    #
    #
    #  // Variable: model
    #  //
    #  // Block abstraction this sequence executes on, defined only when this
    #  // sequence is a user-defined test sequence.
    #  //
    #  uvm_reg_block model;

    #  // Variable: adapter
    #  //
    #  // Adapter to use for translating between abstract register transactions
    #  // and physical bus transactions, defined only when this sequence is a
    #  // translation sequence.
    #  //
    #  uvm_reg_adapter adapter;


    #  // Variable: reg_seqr
    #  //
    #  // Layered upstream "register" sequencer.
    #  //
    #  // Specifies the upstream sequencer between abstract register transactions
    #  // and physical bus transactions. Defined only when this sequence is a
    #  // translation sequence, and we want to "pull" from an upstream sequencer.
    #  //
    #  uvm_sequencer #(uvm_reg_item) reg_seqr;

    def __init__(self, name="uvm_reg_sequence_inst"):
        """         
          Function: new
         
          Create a new instance, giving it the optional `name`.
         
        Args:
            name: 
        """
        super().__init__(name)
        self.reg_seqr = None
        self.adapter = None
        self.model = None
        self.parent_select = LOCAL
        self.upstream_parent = None

    async def body(self):
        """         
          Task: body
         
          Continually gets a register transaction from the configured upstream
          sequencer, `reg_seqr`, and executes the corresponding bus transaction
          via `do_reg_item`.
         
          User-defined RegModel test sequences must override body() and not call
          super.body(), else a warning will be issued and the calling process
          not return.
         
        """
        if self.m_sequencer is None:
            uvm_fatal("NO_SEQR", "Sequence executing as translation sequence, " +
               "but is not associated with a sequencer (m_sequencer == null)")

        if self.reg_seqr is None:
            uvm_warning("REG_XLATE_NO_SEQR",
               "Executing RegModel translation sequence on sequencer " +
               self.m_sequencer.get_full_name() + "' does not have an upstream sequencer defined. "
               + "Execution of register items available only via direct calls to 'do_reg_item'")
            #wait(0);
            await uvm_zero_delay()
        uvm_info("REG_XLATE_SEQ_START",
           "Starting RegModel translation sequence on sequencer " +
           self.m_sequencer.get_full_name() + "'", UVM_LOW)
        while True:
            reg_item = []  # uvm_reg_item reg_item;
            await self.reg_seqr.peek(reg_item)
            await self.do_reg_item(reg_item[0])
            await self.reg_seqr.get(reg_item)
            #0;
            await uvm_zero_delay()


    async def do_reg_item(self, rw):
        """         
          Function: do_reg_item
         
          Executes the given register transaction, `rw`, via the sequencer on
          which this sequence was started (i.e. m_sequencer). Uses the configured
          `adapter` to convert the register transaction into the type expected by
          this sequencer.
         
         virtual task do_reg_item(uvm_reg_item rw);
        Args:
            rw: 
        """
        rws = rw.convert2string()
        if self.m_sequencer is None:
            uvm_fatal("REG/DO_ITEM/NULL","do_reg_item: m_sequencer is null")
        if self.adapter is None:
            uvm_fatal("REG/DO_ITEM/NULL","do_reg_item: adapter handle is null")

        uvm_info("DO_RW_ACCESS","Doing transaction: " + rws, UVM_HIGH)
        #
        if (self.parent_select == LOCAL):
            self.upstream_parent = rw.parent
            rw.parent = self

        if (rw.kind == UVM_WRITE):
            await rw.local_map.do_bus_write(rw, self.m_sequencer, self.adapter)
        else:
            await rw.local_map.do_bus_read(rw, self.m_sequencer, self.adapter)

        if (self.parent_select == LOCAL):
            rw.parent = self.upstream_parent


    #   //----------------------------------
    #   // Group: Convenience Write/Read API
    #   //----------------------------------
    #   //
    #   // The following methods delegate to the corresponding method in the
    #   // register or memory element. They allow a sequence ~body()~ to do
    #   // reads and writes without having to explicitly supply itself to
    #   // ~parent~ sequence argument. Thus, a register write
    #   //
    #   //| model.regA.write(status, value, .parent(this));
    #   //
    #   // can be written instead as
    #   //
    #   //| write_reg(model.regA, status, value);
    #   //


    #   // Task: write_reg
    #   //
    #   // Writes the given register ~rg~ using <uvm_reg::write>, supplying 'this' as
    #   // the ~parent~ argument. Thus,
    #   //
    #   //| write_reg(model.regA, status, value);
    #   //
    #   // is equivalent to
    #   //
    #   //| model.regA.write(status, value, .parent(this));
    #   //
    #   virtual task write_reg(input  uvm_reg           rg,
    #                          output uvm_status_e      status,
    #                          input  uvm_reg_data_t    value,
    #                          input  uvm_path_e        path = UVM_DEFAULT_PATH,
    #                          input  uvm_reg_map       map = null,
    #                          input  int               prior = -1,
    #                          input  uvm_object        extension = null,
    #                          input  string            fname = "",
    #                          input  int               lineno = 0);
    #      if (rg == null)
    #        `uvm_error("NO_REG","Register argument is null")
    #      else
    #        rg.write(status,value,path,map,this,prior,extension,fname,lineno);
    #   endtask


    #   // Task: read_reg
    #   //
    #   // Reads the given register ~rg~ using <uvm_reg::read>, supplying 'this' as
    #   // the ~parent~ argument. Thus,
    #   //
    #   //| read_reg(model.regA, status, value);
    #   //
    #   // is equivalent to
    #   //
    #   //| model.regA.read(status, value, .parent(this));
    #   //
    #   //
    #   virtual task read_reg(input  uvm_reg           rg,
    #                         output uvm_status_e      status,
    #                         output uvm_reg_data_t    value,
    #                         input  uvm_path_e        path = UVM_DEFAULT_PATH,
    #                         input  uvm_reg_map       map = null,
    #                         input  int               prior = -1,
    #                         input  uvm_object        extension = null,
    #                         input  string            fname = "",
    #                         input  int               lineno = 0);
    #      if (rg == null)
    #        `uvm_error("NO_REG","Register argument is null")
    #      else
    #        rg.read(status,value,path,map,this,prior,extension,fname,lineno);
    #   endtask

    #   // Task: poke_reg
    #   //
    #   // Pokes the given register ~rg~ using <uvm_reg::poke>, supplying 'this' as
    #   // the ~parent~ argument. Thus,
    #   //
    #   //| poke_reg(model.regA, status, value);
    #   //
    #   // is equivalent to
    #   //
    #   //| model.regA.poke(status, value, .parent(this));
    #   //
    #   //
    #   virtual task poke_reg(input  uvm_reg           rg,
    #                         output uvm_status_e      status,
    #                         input  uvm_reg_data_t    value,
    #                         input  string            kind = "",
    #                         input  uvm_object        extension = null,
    #                         input  string            fname = "",
    #                         input  int               lineno = 0);
    #      if (rg == null)
    #        `uvm_error("NO_REG","Register argument is null")
    #      else
    #        rg.poke(status,value,kind,this,extension,fname,lineno);
    #   endtask

    #   // Task: peek_reg
    #   //
    #   // Peeks the given register ~rg~ using <uvm_reg::peek>, supplying 'this' as
    #   // the ~parent~ argument. Thus,
    #   //
    #   //| peek_reg(model.regA, status, value);
    #   //
    #   // is equivalent to
    #   //
    #   //| model.regA.peek(status, value, .parent(this));
    #   //
    #   virtual task peek_reg(input  uvm_reg           rg,
    #                         output uvm_status_e      status,
    #                         output uvm_reg_data_t    value,
    #                         input  string            kind = "",
    #                         input  uvm_object        extension = null,
    #                         input  string            fname = "",
    #                         input  int               lineno = 0);
    #      if (rg == null)
    #        `uvm_error("NO_REG","Register argument is null")
    #      else
    #        rg.peek(status,value,kind,this,extension,fname,lineno);
    #   endtask

    #   // Task: update_reg
    #   //
    #   // Updates the given register ~rg~ using <uvm_reg::update>, supplying 'this' as
    #   // the ~parent~ argument. Thus,
    #   //
    #   //| update_reg(model.regA, status, value);
    #   //
    #   // is equivalent to
    #   //
    #   //| model.regA.update(status, value, .parent(this));
    #   //
    #   virtual task update_reg(input  uvm_reg           rg,
    #                           output uvm_status_e      status,
    #                           input  uvm_path_e        path = UVM_DEFAULT_PATH,
    #                           input  uvm_reg_map       map = null,
    #                           input  int               prior = -1,
    #                           input  uvm_object        extension = null,
    #                           input  string            fname = "",
    #                           input  int               lineno = 0);
    #      if (rg == null)
    #        `uvm_error("NO_REG","Register argument is null")
    #      else
    #        rg.update(status,path,map,this,prior,extension,fname,lineno);
    #   endtask


    #   // Task: mirror_reg
    #   //
    #   // Mirrors the given register ~rg~ using <uvm_reg::mirror>, supplying 'this' as
    #   // the ~parent~ argument. Thus,
    #   //
    #   //| mirror_reg(model.regA, status, UVM_CHECK);
    #   //
    #   // is equivalent to
    #   //
    #   //| model.regA.mirror(status, UVM_CHECK, .parent(this));
    #   //
    #   virtual task mirror_reg(input  uvm_reg       rg,
    #                           output uvm_status_e  status,
    #                           input  uvm_check_e   check  = UVM_NO_CHECK,
    #                           input  uvm_path_e    path = UVM_DEFAULT_PATH,
    #                           input  uvm_reg_map   map = null,
    #                           input  int           prior = -1,
    #                           input  uvm_object    extension = null,
    #                           input  string        fname = "",
    #                           input  int           lineno = 0);
    #      if (rg == null)
    #        `uvm_error("NO_REG","Register argument is null")
    #      else
    #        rg.mirror(status,check,path,map,this,prior,extension,fname,lineno);
    #   endtask

    #   // Task: write_mem
    #   //
    #   // Writes the given memory ~mem~ using <uvm_mem::write>, supplying 'this' as
    #   // the ~parent~ argument. Thus,
    #   //
    #   //| write_mem(model.regA, status, offset, value);
    #   //
    #   // is equivalent to
    #   //
    #   //| model.regA.write(status, offset, value, .parent(this));
    #   //
    #   virtual task write_mem(input  uvm_mem           mem,
    #                          output uvm_status_e      status,
    #                          input  uvm_reg_addr_t    offset,
    #                          input  uvm_reg_data_t    value,
    #                          input  uvm_path_e        path = UVM_DEFAULT_PATH,
    #                          input  uvm_reg_map       map = null,
    #                          input  int               prior = -1,
    #                          input  uvm_object        extension = null,
    #                          input  string            fname = "",
    #                          input  int               lineno = 0);
    #      if (mem == null)
    #        `uvm_error("NO_MEM","Memory argument is null")
    #      else
    #        mem.write(status,offset,value,path,map,this,prior,extension,fname,lineno);
    #   endtask

    #   // Task: read_mem
    #   //
    #   // Reads the given memory ~mem~ using <uvm_mem::read>, supplying 'this' as
    #   // the ~parent~ argument. Thus,
    #   //
    #   //| read_mem(model.regA, status, offset, value);
    #   //
    #   // is equivalent to
    #   //
    #   //| model.regA.read(status, offset, value, .parent(this));
    #   //
    #   //
    #   virtual task read_mem(input  uvm_mem           mem,
    #                         output uvm_status_e      status,
    #                         input  uvm_reg_addr_t    offset,
    #                         output uvm_reg_data_t    value,
    #                         input  uvm_path_e        path = UVM_DEFAULT_PATH,
    #                         input  uvm_reg_map       map = null,
    #                         input  int               prior = -1,
    #                         input  uvm_object        extension = null,
    #                         input  string            fname = "",
    #                         input  int               lineno = 0);
    #      if (mem == null)
    #        `uvm_error("NO_MEM","Memory argument is null")
    #      else
    #        mem.read(status,offset,value,path,map,this,prior,extension,fname,lineno);
    #   endtask

    #   // Task: poke_mem
    #   //
    #   // Pokes the given memory ~mem~ using <uvm_mem::poke>, supplying 'this' as
    #   // the ~parent~ argument. Thus,
    #   //
    #   //| poke_mem(model.regA, status, offset, value);
    #   //
    #   // is equivalent to
    #   //
    #   //| model.regA.poke(status, offset, value, .parent(this));
    #   //
    #   //
    #   virtual task poke_mem(input  uvm_mem           mem,
    #                         output uvm_status_e      status,
    #                         input  uvm_reg_addr_t    offset,
    #                         input  uvm_reg_data_t    value,
    #                         input  string            kind = "",
    #                         input  uvm_object        extension = null,
    #                         input  string            fname = "",
    #                         input  int               lineno = 0);
    #      if (mem == null)
    #        `uvm_error("NO_MEM","Memory argument is null")
    #      else
    #        mem.poke(status,offset,value,kind,this,extension,fname,lineno);
    #   endtask

    #   // Task: peek_mem
    #   //
    #   // Peeks the given memory ~mem~ using <uvm_mem::peek>, supplying 'this' as
    #   // the ~parent~ argument. Thus,
    #   //
    #   //| peek_mem(model.regA, status, offset, value);
    #   //
    #   // is equivalent to
    #   //
    #   //| model.regA.peek(status, offset, value, .parent(this));
    #   //
    #   virtual task peek_mem(input  uvm_mem           mem,
    #                         output uvm_status_e      status,
    #                         input  uvm_reg_addr_t    offset,
    #                         output uvm_reg_data_t    value,
    #                         input  string            kind = "",
    #                         input  uvm_object        extension = null,
    #                         input  string            fname = "",
    #                         input  int               lineno = 0);
    #      if (mem == null)
    #        `uvm_error("NO_MEM","Memory argument is null")
    #      else
    #        mem.peek(status,offset,value,kind,this,extension,fname,lineno);
    #   endtask

    #  // Function- put_response
    #  //
    #  // not user visible. Needed to populate this sequence's response
    #  // queue with any bus item type.
    #  //
    #  virtual function void put_response(uvm_sequence_item response_item);
    #    put_base_response(response_item);
    #  endfunction
    #
    #endclass
uvm_object_utils(UVMRegSequence)


#//------------------------------------------------------------------------------
#// Class: UVMRegFrontdoor
#//
#// Facade class for register and memory frontdoor access.
#//------------------------------------------------------------------------------
#//
#// User-defined frontdoor access sequence
#//
#// Base class for user-defined access to register and memory reads and writes
#// through a physical interface.
#//
#// By default, different registers and memories are mapped to different
#// addresses in the address space and are accessed via those exclusively
#// through physical addresses.
#//
#// The frontdoor allows access using a non-linear and/or non-mapped mechanism.
#// Users can extend this class to provide the physical access to these registers.
#//


class UVMRegFrontdoor(UVMRegSequence):  # (uvm_sequence #(uvm_sequence_item));

    def __init__(self, name="reg_frontdoor_seq"):
        """         
           Function: new
          
           Constructor, new object given optional `name`.
          
        Args:
            name: 
        """
        super().__init__(name)
        #
        #   // Variable: rw_info
        #   //
        #   // Holds information about the register being read or written
        #   //
        #   uvm_reg_item rw_info;
        self.rw_info = None
        self.fname = ""
        self.lineno = 0
        #
        #   // Variable: sequencer
        #   //
        #   // Sequencer executing the operation
        #   //
        #   uvm_sequencer_base sequencer;
        self.sequencer = None



#//----------------------------------------------------------------------
#//   Copyright 2010 Mentor Graphics Corporation
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2020 Matthew Ballance
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.base.uvm_port_base import *
from ..tlm1 import (UVM_PORT_COMMON)
from .uvm_tlm2_defines import UVM_TLM_B_MASK
from .uvm_tlm2_imps import (UVM_TLM_B_TRANSPORT_IMP)

#//----------------------------------------------------------------------
#// Title -- NODOCS -- TLM Socket Base Classes
#//
#// A collection of base classes, one for each socket type.  The reason
#// for having a base class for each socket is that all the socket (base)
#// types must be known before connect is defined.  Socket connection
#// semantics are provided in the derived classes, which are user
#// visible.
#//
#// Termination Sockets - A termination socket must be the terminus
#// of every TLM path.  A transaction originates with an initiator socket
#// and ultimately ends up in a target socket.  There may be zero or more
#// pass-through sockets between initiator and target.
#//
#// Pass-through Sockets - Pass-through initiators are ports and contain
#// exports for instance IS-A port and HAS-A export. Pass-through targets
#// are the opposite, they are exports and contain ports.
#//----------------------------------------------------------------------


#//----------------------------------------------------------------------
#// Class -- NODOCS -- uvm_tlm_b_target_socket_base
#//
#// IS-A forward imp; has no backward path except via the payload
#// contents.
#//----------------------------------------------------------------------
from uvm.base.uvm_port_base import UVMPortBase
from uvm.base.uvm_object_globals import UVM_IMPLEMENTATION, UVM_PORT, UVM_EXPORT
from uvm.tlm1.uvm_tlm_imps import UVM_TLM_GET_TYPE_NAME, UVM_PORT_COMMON,\
    UVM_EXPORT_COMMON
from uvm.tlm2.uvm_tlm2_defines import UVM_TLM_B_MASK, UVM_TLM_NB_FW_MASK
from uvm.tlm2.uvm_tlm2_imps import UVM_TLM_B_TRANSPORT_IMP,\
    UVM_TLM_NB_TRANSPORT_BW_IMP, UVM_TLM_NB_TRANSPORT_FW_IMP
from uvm.tlm2.uvm_tlm2_ports import UVMTLMNbTransportBwPort
from uvm.tlm2.uvm_tlm2_exports import UVMTLMNbTransportBwExport

class UVMTLMBTargetSocketBase(UVMPortBase):

    def __init__(self, name, parent):
        super().__init__(name, parent, UVM_IMPLEMENTATION, 1, 1)
        self.m_if_mask = UVM_TLM_B_MASK

UVM_TLM_GET_TYPE_NAME(UVMTLMBTargetSocketBase)
    
#
#//----------------------------------------------------------------------
#// Class: uvm_tlm_b_initiator_socket_base
#//
#// IS-A forward port; has no backward path except via the payload
#// contents
#//----------------------------------------------------------------------
class UVMTLMBInitiatorSocketBase(UVMPortBase):
    pass

UVMTLMBInitiatorSocketBase = UVM_PORT_COMMON(UVMTLMBInitiatorSocketBase, UVM_TLM_B_MASK, "UVMTLMBInitiatorSocketBase")
UVM_TLM_B_TRANSPORT_IMP("m_if", UVMTLMBInitiatorSocketBase)

#
#//----------------------------------------------------------------------
#// Class: uvm_tlm_nb_target_socket_base
#//
#// IS-A forward imp; HAS-A backward port
#//----------------------------------------------------------------------
class UVMTLMNbTargetSocketBase(UVMPortBase):

    def __init__(self, name, parent):
        super().__init__(name, parent, UVM_IMPLEMENTATION, 1, 1)
        self.m_if_mask = UVM_TLM_NB_FW_MASK
#        self.bw_port = UVMTLMNbTransportBwPort(name, parent, port_type, min_size, max_size)None # TODO: uvm_tlm_nb_transport_bw_port

UVM_TLM_GET_TYPE_NAME(UVMTLMNbTargetSocketBase)
UVM_TLM_NB_TRANSPORT_BW_IMP('bw_port', UVMTLMNbTargetSocketBase)


#//----------------------------------------------------------------------
#// Class: uvm_tlm_nb_initiator_socket_base
#//
#// IS-A forward port; HAS-A backward imp
#//----------------------------------------------------------------------
class UVMTLMNbInitiatorSocketBase(UVMPortBase):
    
    def __init__(self, name, parent):
        super().__init__(name, parent, UVM_PORT, 1, 1)
        self.m_if_mask = UVM_TLM_NB_FW_MASK

UVM_TLM_GET_TYPE_NAME(UVMTLMNbInitiatorSocketBase)
UVM_TLM_NB_TRANSPORT_FW_IMP('m_if', UVMTLMNbInitiatorSocketBase)

#//----------------------------------------------------------------------
#// Class -- NODOCS -- uvm_tlm_nb_passthrough_initiator_socket_base
#//
#// IS-A forward port; HAS-A backward export
#//----------------------------------------------------------------------
class UVMTLMNbPassthroughInitiatorSocketBase(UVMPortBase):

    def __init__(self, name, parent, min_size=1, max_size=1):
        super().__init__(name, parent, UVM_PORT, min_size, max_size)
        self.m_if_mask = UVM_TLM_NB_FW_MASK
        self.bw_export = UVMTLMNbTransportBwExport("bw_export", self.get_comp())

UVM_TLM_GET_TYPE_NAME(UVMTLMNbPassthroughInitiatorSocketBase)
UVM_TLM_NB_TRANSPORT_FW_IMP('m_if', UVMTLMNbPassthroughInitiatorSocketBase)
UVM_TLM_NB_TRANSPORT_BW_IMP('bw_export', UVMTLMNbPassthroughInitiatorSocketBase)

#//----------------------------------------------------------------------
#// Class: uvm_tlm_nb_passthrough_target_socket_base
#//
#// IS-A forward export; HAS-A backward port
#//----------------------------------------------------------------------
class UVMTLMNbPassthroughTargetSocketBase(UVMPortBase):

    def __init__(self, name, parent, min_size=1, max_size=1):
        super().__init__(name, parent, UVM_EXPORT, min_size, max_size)
        self.m_if_mask = UVM_TLM_NB_FW_MASK
        self.bw_port = UVMTLMNbTransportBwPort("bw_port", self.get_comp())

UVM_TLM_GET_TYPE_NAME(UVMTLMNbPassthroughTargetSocketBase)
UVM_TLM_NB_TRANSPORT_FW_IMP('m_if', UVMTLMNbPassthroughTargetSocketBase)
UVM_TLM_NB_TRANSPORT_BW_IMP('bw_port', UVMTLMNbPassthroughTargetSocketBase)

#//----------------------------------------------------------------------
#// Class: uvm_tlm_b_passthrough_initiator_socket_base
#//
#// IS-A forward port
#//----------------------------------------------------------------------
class UVMTLMBPassthroughInitiatorSocketBase(UVMPortBase):
    pass

UVMTLMBPassthroughInitiatorSocketBase = UVM_PORT_COMMON(UVMTLMBPassthroughInitiatorSocketBase, UVM_TLM_B_MASK, "uvm_tlm_b_passthrough_initiator_socket")
UVM_TLM_B_TRANSPORT_IMP('m_if', UVMTLMBPassthroughInitiatorSocketBase)

#//----------------------------------------------------------------------
#// Class -- NODOCS -- uvm_tlm_b_passthrough_target_socket_base
#//
#// IS-A forward export
#//----------------------------------------------------------------------
class UVMTLMBPassthroughTargetSocketBase(UVMPortBase):
    pass

UVMTLMBPassthroughTargetSocketBase = UVM_EXPORT_COMMON(UVMTLMBPassthroughTargetSocketBase, UVM_TLM_B_MASK, "uvm_tlm_b_passthrough_target_socket")
UVM_TLM_B_TRANSPORT_IMP('m_if', UVMTLMBPassthroughTargetSocketBase)




#
#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

# About: UVM Reporting methods
# This example will illustrate the usage of UVM reporting functions,
# and mainly focusing on the fine-tuning of reporting

# To get more details about the reporting related methods, check the file:
#   - uvm/src/base/uvm_report_object.py
#   - uvm/src/base/uvm_component.py  - Hierarchical reporting methods

import cocotb
from cocotb.triggers import Timer

from uvm.macros import uvm_component_utils, uvm_info
from uvm import (UVMCoreService, UVMComponent, UVM_MEDIUM, sv, UVMTest,
    run_test, UVM_LOW, UVM_LOG, UVM_DISPLAY, UVM_INFO, UVMReportCatcher)


class my_child(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.tag = "MY_CHILD"

    async def main_phase(self, phase):
        uvm_info(self.tag, 'main_phase started in child', UVM_MEDIUM)

    def get_packet(self):
        uvm_info("PKTGEN", sv.sformatf("Getting a packet from %s (%s)",
            self.get_full_name(), self.get_type_name()), UVM_MEDIUM)
        return super().get_packet()


# Use the macro in a class to implement factory registration along with other
# utilities (create, get_type_name). To just do factory registration, use the
# macro `uvm_object_registry(mygen,"mygen")
uvm_component_utils(my_child)


class my_top_test(UVMTest):

    def __init__(self, name="my_top_test", parent=None):
        super().__init__(name, parent)
        self.tag = 'MY_TOP_TEST'

    def build_phase(self, phase):
        super().build_phase(phase)
        self.child = my_child.type_id.create('my_child', self)

    def connect_phase(self, phase):
        self.def_file = open('uvm_master_log.log', 'w')
        self.set_report_default_file_hier(self.def_file)
        self.set_report_severity_action_hier(UVM_INFO, UVM_LOG | UVM_DISPLAY)

    async def main_phase(self, phase):
        phase.raise_objection(self, 'main_phase_started')
        uvm_info(self.tag, 'main_phase started', UVM_MEDIUM)
        await Timer(10, 'NS')
        phase.drop_objection(self, 'main_phase_ending')

    def final_phase(self, phase):
        #uvm_info(self.tag, "Closing the file handle now", UVM_LOW)
        self.def_file.close()
        uvm_info(self.tag, "Closing the file handle now", UVM_LOW)


# Use the macro in after the class to implement factory registration along with other
# utilities (create, get_type_name).
uvm_component_utils(my_top_test)


@cocotb.test()
async def module_top(dut):
    # cs_ = UVMCoreService.get()  # type: UVMCoreService
    # uvm_root = cs_.get_root()
    # factory = cs_.get_factory()
    # factory.print_factory(1)

    # If a string is used to run_test, run_test will used the string based factory
    # create method to create an object of the desired type.
    await run_test("my_top_test")



#//
#//-----------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2013 NVIDIA Corporation
#//   Copyright 2020 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//-----------------------------------------------------------------------------

from ..dap.uvm_simple_lock_dap import uvm_simple_lock_dap
from ..base.uvm_object import UVMObject
from ..base.sv import sv
from ..macros.uvm_object_defines import uvm_object_utils
from ..macros.uvm_message_defines import uvm_warning
from .uvm_tr_stream import UVMTextTrStream, UVMTrStream
from .uvm_recorder import UVMRecorder
from .uvm_links import (UVMParentChildLink, UVMRelatedLink)

NO_FILE_OPEN = 0

#//------------------------------------------------------------------------------
#// File: Transaction Recording Databases
#//
#// The UVM "Transaction Recording Database" classes are an abstract representation
#// of the backend tool which is recording information for the user.  Usually this
#// tool would be dumping information such that it can be viewed with the ~waves~
#// of the DUT.
#//


#//------------------------------------------------------------------------------
#//
#// CLASS: UVMTrDatabase
#//
#// The ~UVMTrDatabase~ class is intended to hide the underlying database implementation
#// from the end user, as these details are often vendor or tool-specific.
#//
#// The ~UVMTrDatabase~ class is pure virtual, and must be extended with an
#// implementation.  A default text-based implementation is provided via the
#// <UVMTextTrDatabase> class.
#//

class UVMTrDatabase(UVMObject):
    #
    #   // Variable- m_streams

    def __init__(self, name="unnamed-UVMTrDatabase"):
        """
           Function: new
           Constructor

           Parameters:
           name - Instance name
        Args:
            name:
        """
        UVMObject.__init__(self, name)
        # Tracks the opened state of the database
        self.m_is_opened = False
        # Used for tracking streams which are between the open and closed states
        self.m_streams = {}  # bit m_streams[UVMTrStream]

    def open_db(self):
        """
           Group: Database API

           Function: open_db
           Open the backend connection to the database.

           If the database is already open, then this
           method will return 1.

           Otherwise, the method will call `do_open_db`,
           and return the result.
        Returns:
        """
        if self.m_is_opened is False:
            self.m_is_opened = self.do_open_db()
        return self.m_is_opened

    def close_db(self):
        """

           Function: close_db
           Closes the backend connection to the database.

           Closing a database implicitly closes and
           frees all `uvm_tr_streams` within the database.

           If the database is already closed, then this
           method will return 1.

           Otherwise, this method will trigger a `do_close_db`
           call, and return the result.
        Returns:
        """
        if self.m_is_opened:
            if self.do_close_db():
                self.m_is_opened = 0
        return self.m_is_opened == 0

    def is_open(self):
        """
           Function: is_open
           Returns the open/closed status of the database.

           This method returns 1 if the database has been
           successfully opened, but not yet closed.

        Returns:
        """
        return self.m_is_opened


    #   // Group: Stream API
    #

    def open_stream(self, name, scope="", type_name=""):
        """
           Function: open_stream
           Provides a reference to a `stream` within the
           database.

           Parameters:
             name - A string name for the stream.  This is the name associated
                    with the stream in the database.
             scope - An optional scope for the stream.
             type_name - An optional name describing the type of records which
                         will be created in this stream.

           The method returns a reference to a `UVMTrStream`
           object if successful, `null` otherwise.

           This method will trigger a `do_open_stream` call, and if a
           non `null` stream is returned, then <UVMTrStream::do_open>
           will be called.

           Streams can only be opened if the database is
           open (per `is_open`).  Otherwise the request will
           be ignored, and `null` will be returned.
          function UVMTrStream open_stream(string name,
                                             string scope="",
                                             string type_name="")
        Args:
            name:
            scope:
            type_name:
        Returns:
        """
        if not self.open_db():
            return None
        else:
            #process p = process::self()
            p = None
            s = ""

            if p is not None:
                s = p.get_randstate()

            open_stream = self.do_open_stream(name, scope, type_name)

            if open_stream is not None:
                self.m_streams[open_stream] = 1
                open_stream.m_do_open(self, scope, type_name)

            if p is not None:
                p.set_randstate(s)
            return open_stream


    def m_free_stream(self, stream):
        """
           Function- m_free_stream
           Removes stream from the internal array
        Args:
            stream:
        """
        if stream in self.m_streams:
            del self.m_streams[stream]
    #   endfunction : m_free_stream

    #
    #   // Function: get_streams
    #   // Provides a queue of all streams within the database.
    #   //
    #   // Parameters:
    #   // q - A reference to a queue of <UVMTrStream>s
    #   //
    #   // The ~get_streams~ method returns the size of the queue,
    #   // such that the user can conditionally process the elements.
    #   //
    #   // | UVMTrStream stream_q[$]
    #   // | if (my_db.get_streams(stream_q)) begin
    #   // |   // Process the queue...
    #   // | end
    #   function unsigned get_streams(ref UVMTrStream q[$])
    #      // Clear out the queue first...
    #      q.delete()
    #      // Then fill in the values
    #      foreach (m_streams[idx])
    #        q.push_back(idx)
    #      // Finally, return the size of the queue
    #      return q.size()
    #   endfunction : get_streams
    #
    #   // Group: Link API

    def establish_link(self, link):
        """
           Function: establish_link
           Establishes a `link` between two elements in the database

           Links are only supported between `streams` and `records`
           within a single database.

           This method will trigger a `do_establish_link` call.
        Args:
            link:
        """
        s_lhs = []  # UVMTrStream
        s_rhs = []  # UVMTrStream
        r_lhs = []  # uvm_recorder
        r_rhs = []  # uvm_recorder
        lhs = link.get_lhs()
        rhs = link.get_rhs()
        db = None  # UVMTrDatabase

        if lhs is None:
            uvm_warning("UVM/TR_DB/BAD_LINK",
                "left hand side '<None>' is not supported in links for 'UVMTrDatabase'")

        if rhs is None:
            uvm_warning("UVM/TR_DB/BAD_LINK",
                "right hand side '<None>' is not supported in links for 'UVMTrDatabase'")
            return


        if (not sv.cast(s_lhs, lhs, UVMTrStream)) or (not sv.cast(r_lhs, lhs,
            UVMRecorder)):
            uvm_warning("UVM/TR_DB/BAD_LINK",
                sv.sformatf("left hand side of type '%s' not supported in links for 'UVMTrDatabase'",
                lhs.get_type_name()))
            return
        else:
            s_lhs = s_lhs[0]
            r_lhs = r_lhs[0]

        if (not sv.cast(s_rhs, rhs, UVMTrStream)) or (not sv.cast(r_rhs, rhs, UVMRecorder)):
            uvm_warning("UVM/TR_DB/BAD_LINK",
                sv.sformatf("right hand side of type '%s' not supported in links for 'uvm_record_datbasae'",
                rhs.get_type_name()))
            return
        else:
            s_rhs = s_rhs[0]
            r_rhs = r_rhs[0]

        if r_lhs is not None:
            s_lhs = r_lhs.get_stream()

        if r_rhs is not None:
            s_rhs = r_rhs.get_stream()


        if ((s_lhs is not None) and (s_lhs.get_db() != self)):
            db = s_lhs.get_db()
            uvm_warning("UVM/TR_DB/BAD_LINK",
                sv.sformatf("attempt to link stream from '%s' into '%s'",
                db.get_name(), self.get_name()))
            return

        if ((s_rhs is not None) and (s_rhs.get_db() != self)):
            db = s_rhs.get_db()
            uvm_warning("UVM/TR_DB/BAD_LINK",
                sv.sformatf("attempt to link stream from '%s' into '%s'",
                db.get_name(), self.get_name()))
            return


        self.do_establish_link(link)
        #   endfunction : establish_link

    #
    #   // Group: Implementation Agnostic API
    #   //
    #
    #   // Function: do_open_db
    #   // Backend implementation of <open_db>
    def do_open_db(self):
        raise Exception('do_open_db is pure virtual function')

    #   // Function: do_close_db
    #   // Backend implementation of <close_db>
    #   pure virtual protected function bit do_close_db()
    #
    #   // Function: do_open_stream
    #   // Backend implementation of <open_stream>
    #   pure virtual protected function UVMTrStream do_open_stream(string name,
    #                                                                string scope,
    #                                                                string type_name)
    #
    #   // Function: do_establish_link
    #   // Backend implementation of <establish_link>
    #   pure virtual protected function void do_establish_link(uvm_link_base link)
    def do_establish_link(self, link):
        raise Exception('do_establish_link is pure virtual function')



class UVMTextTrDatabase(UVMTrDatabase):
    """
    CLASS: UVMTextTrDatabase

    The ~UVMTextTrDatabase~ is the default implementation for the
    <UVMTrDatabase>.  It provides the ability to store recording information
    into a textual log file.
    """


    #   // Variable- m_filename_dap
    #   // Data Access Protected Filename
    #   local uvm_simple_lock_dap#(string) m_filename_dap
    #
    #   // Variable- m_file
    #   UVM_FILE m_file

    def __init__(self, name="unnamed-UVMTextTrDatabase"):
        """
           Function: new
           Constructor

           Parameters:
           name - Instance name
        Args:
            name:
        """
        UVMTrDatabase.__init__(self, name)
        self.m_filename_dap = uvm_simple_lock_dap("filename_dap")
        self.m_filename_dap.set("tr_db.log")
        self.m_file = NO_FILE_OPEN

    def do_open_db(self) -> bool:
        """
           Group: Implementation Agnostic API

           Function: do_open_db
           Open the backend connection to the database.

           Text-Backend implementation of <UVMTrDatabase::open_db>.

           The text-backend will open a text file to dump all records in to.  The name
           of this text file is controlled via `set_file_name`.

           This will also lock the `file_name`, so that it cannot be
           modified while the connection is open.
        Returns:
        """
        if self.m_file == NO_FILE_OPEN:
            self.m_file = sv.fopen(self.m_filename_dap.get(), "a")
            if self.m_file != NO_FILE_OPEN:
                self.m_filename_dap.lock()
        return (self.m_file != NO_FILE_OPEN)

    def do_close_db(self) -> int:
        """

           Function: do_close_db
           Close the backend connection to the database.

           Text-Backend implementation of <UVMTrDatabase::close_db>.

           The text-backend will close the text file used to dump all records in to,
           if it is currently opened.

           This unlocks the `file_name`, allowing it to be modified again.
        Returns:
        """
        if self.m_file != NO_FILE_OPEN:
            #fork // Needed because $fclose is a task
            sv.fclose(self.m_file)
            #join_none
            self.m_filename_dap.unlock()
        return 1

    def do_open_stream(self, name: str, scope: str, typename: str) -> UVMTextTrStream:
        """
           Function: do_open_stream
           Provides a reference to a `stream` within the
           database.

           Text-Backend implementation of <UVMTrDatabase::open_stream>
          protected virtual function UVMTrStream do_open_stream(string name,
                                                                  string scope,
                                                                  string type_name)
        Args:
            name:
            scope:
            typename:
        Returns:
        """
        # puvm_text_tr_stream
        m_stream = UVMTextTrStream.type_id.create(name)
        return m_stream


    def do_establish_link(self, link) -> None:
        """
           Function: do_establish_link
           Establishes a `link` between two elements in the database

           Text-Backend implementation of <UVMTrDatabase::establish_link>.
        Args:
            link:
        """
        cast_r_lhs = []
        cast_r_rhs = []  # uvm_recorder
        lhs = link.get_lhs()
        rhs = link.get_rhs()

        sv.cast(cast_r_lhs, lhs, UVMRecorder)
        sv.cast(cast_r_rhs, rhs, UVMRecorder)
        r_lhs = cast_r_lhs[0]
        r_rhs = cast_r_rhs[0]

        if (r_lhs is None or r_rhs is None):
            return
        else:
            pc_link = []  # uvm_parent_child_link
            re_link = []  # uvm_related_link
            if sv.cast(pc_link, link, UVMParentChildLink):
                sv.fdisplay(self.m_file, "  LINK @%0t {{TXH1:%0d TXH2:%0d RELATION=%0s}}",
                    sv.time(), r_lhs.get_handle(), r_rhs.get_handle(), "child")

            elif sv.cast(re_link, link, UVMRelatedLink):
                sv.fdisplay(self.m_file, "  LINK @%0t {{TXH1:%0d TXH2:%0d RELATION=%0s}}",
                    sv.time(), r_lhs.get_handle(), r_rhs.get_handle(), "")


    #   // Group: Implementation Specific API
    #
    #   // Function: set_file_name
    #   // Sets the file name which will be used for output.
    #   //
    #   // The ~set_file_name~ method can only be called prior to ~open_db~.
    #   //
    #   // By default, the database will use a file named "tr_db.log".
    def set_file_name(self, filename: str) -> None:
        if filename == "":
            uvm_warning("UVM/TXT_DB/EMPTY_NAME",
                       "Ignoring attempt to set file name to ''!")
            return


        if not self.m_filename_dap.try_set(filename):
            uvm_warning("UVM/TXT_DB/SET_AFTER_OPEN",
                        "Ignoring attempt to change file name after opening the db!")
        return


uvm_object_utils(UVMTextTrDatabase)
